#! /usr/bin/env python
# coding: utf-8
#
""" ForecastGA: Models """

MODELS = {
    "ARIMA": {
        "name": "ARIMA",
        "description": "Automated ARIMA Modelling",
        "loc": "forecastga.models.arima",
        "class": "ARIMA_Model",
        "status": "active",
    },
    "Prophet": {
        "name": "Prophet",
        "description": "Modeling Multiple Seasonality With Linear or Non-linear Growth",
        "loc": "forecastga.models.prophet",
        "class": "Prophet_Model",
        "status": "active",
    },
    "ProphetBC": {
        "name": "ProphetBC",
        "description": "Prophet Model with Box-Cox transform of the data",
        "loc": "forecastga.models.prophet_boxcox",
        "class": "Prophet_BoxCox_Model",
        "status": "active",
    },
    "SARIMA": {
        "name": "SARIMA",
        "description": "A seasonal autoregressive integrated moving average (SARIMA) model",
        "loc": "forecastga.models.sarima",
        "class": "SARIMA_Model",
        "status": "not implemented",
    },
    "SARIMAX": {
        "name": "SARIMAX",
        "description": "Example: https://gist.github.com/natzir/befe1ff229fc2d0c01e0411d5fdd5209",
        "loc": "forecastga.models.sarimax",
        "class": "SARIMAX_Model",
        "status": "not implemented",
    },
    "HWAAS": {
        "name": "HWAAS",
        "description": "Exponential Smoothing With Additive Trend and Additive Seasonality",
        "loc": "forecastga.models.hwaas",
        "class": "HWAAS_Model",
        "status": "active",
    },
    "HWAMS": {
        "name": "HWAMS",
        "description": "Exponential Smoothing with Additive Trend and Multiplicative Seasonality",
        "loc": "forecastga.models.hwams",
        "class": "HWAMS_Model",
        "status": "active",
    },
    "NBEATS": {
        "name": "NBEATS",
        "description": "Neural basis expansion analysis (now fixed at 20 Epochs)",
        "loc": "forecastga.models.nbeats",
        "class": "NBEATS_Model",
        "status": "active",
    },
    "Gluonts": {
        "name": "Gluonts",
        "description": "RNN-based Model (now fixed at 20 Epochs)",
        "loc": "forecastga.models.gluonts",
        "class": "Gluonts_Model",
        "status": "active",
    },
    "TATS": {
        "name": "TATS",
        "description": "Seasonal and Trend no Box Cox",
        "loc": "forecastga.models.tats",
        "class": "TATS_Model",
        "status": "active",
    },
    "TBAT": {
        "name": "TBAT",
        "description": "Trend and Box Cox",
        "loc": "forecastga.models.tbat",
        "class": "TBAT_Model",
        "status": "active",
    },
    "TBATS1": {
        "name": "TBATS1",
        "description": "Trend, Seasonal (one), and Box Cox",
        "loc": "forecastga.models.tbats1",
        "class": "TBATS1_Model",
        "status": "active",
    },
    "TBATP1": {
        "name": "TBATP1",
        "description": "TBATS1 but Seasonal Inference is Hardcoded by Periodicity",
        "loc": "forecastga.models.tbatp1",
        "class": "TBATP1_Model",
        "status": "active",
    },
    "TBATS2": {
        "name": "TBATS2",
        "description": "TBATS1 With Two Seasonal Periods",
        "loc": "forecastga.models.tbats2",
        "class": "TBATS2_Model",
        "status": "active",
    },
    "PYAF": {
        "name": "PYAF",
        "description": "PYAF",
        "loc": "forecastga.models.pyaf",
        "class": "PYAF_Model",
        "status": "active but not tested",
    },
}



# encoding: utf-8

import os
import copy
import operator
import functools

from . import date
from .functional import memoize, immutable, identity, soak, vectorize
from .server import single_serve
from .string import format, affix, paste, cut


# Python 2 and 3 compatibility
try:
    basestring = basestring
    unicode = unicode
    input = raw_input
except NameError:
    basestring = str
    unicode = str
    input = input

try:
    import __builtin__ as builtins
    from StringIO import StringIO
except ImportError:
    import builtins
    from io import StringIO


# return a path relative to the package root
def here(*segments):
    current = os.path.dirname(__file__)
    return os.path.realpath(os.path.join(current, "..", *segments))


# flatten nested lists
def flatten(nested_list):
    return functools.reduce(operator.add, nested_list)


# wrap scalars into a list
def wrap(obj):
    if isinstance(obj, list):
        return obj

    return [obj]


# substitute new dictionary keys
def translate(d, mapping):
    d = copy.copy(d)

    for src, dest in mapping.items():
        if src in d:
            d[dest] = d[src]
            del d[src]

    return d


# retain only whitelisted keys in a dictionary
def whitelist(d, allowed):
    return {k: v for k, v in d.items() if k in allowed}


# similar to whitelist, but ordered and returns only values, not keys
def pick(obj, allowed):
    if isinstance(obj, dict):
        get = lambda key: obj[key]
    else:
        get = lambda key: getattr(obj, key)

    values = []
    for key in allowed:
        values.append(get(key))

    return values


# test if an object is falsy or contains only falsy values
def isempty(obj):
    if isinstance(obj, list):
        return not len(list(filter(None, obj))) > 0
    if isinstance(obj, dict):
        return not len(obj) > 0

    return not obj



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: HWAAS Model"""

from statsmodels.tsa.holtwinters import ExponentialSmoothing

from forecastga.models.base import BaseModel


class HWAAS_Model(BaseModel):
    """HWAAS Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        for i in range(2):
            use_boxcox = [True, False][i]
            try:
                self.model = ExponentialSmoothing(
                    self.train_df,
                    seasonal_periods=self.seasons,
                    trend="add",
                    seasonal="add",
                    damped_trend=True,
                ).fit(use_boxcox=use_boxcox)
                break
            except:
                continue

    def forecast(self):
        self.prediction = self.model.forecast(self.forecast_len)



# encoding: utf-8


class GoogleAnalyticsError(Exception):
    pass


class InvalidRequestError(GoogleAnalyticsError):
    # invalid parameter, bad request
    pass


class NotPermittedError(GoogleAnalyticsError):
    # invalid credentials, no permission
    pass


class LimitExceededError(GoogleAnalyticsError):
    # quota, rate limit, ...
    pass


class ServerError(GoogleAnalyticsError):
    # internal server error / backend error
    pass



#! /usr/bin/env python
# coding: utf-8
#

"""Forecastga: Holiday Helpers"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import inspect
import unicodedata

import pandas as pd
import numpy as np

import holidays as hdays_part1
import fbprophet.hdays as hdays_part2
from fbprophet.make_holidays import make_holidays_df


def utf8_to_ascii(text):
    """Holidays often have utf-8 characters. These are not allowed in R
    package data (they generate a NOTE).
    TODO: revisit whether we want to do this lossy conversion.
    """
    ascii_text = (
        unicodedata.normalize("NFD", text)
        .encode("ascii", "ignore")
        .decode("ascii")
        .strip()
    )
    # Check if anything converted
    if sum(1 for x in ascii_text if x not in [" ", "(", ")", ","]) == 0:
        return "FAILED_TO_PARSE"
    else:
        return ascii_text


def generate_holidays_file():
    """Generate csv file of all possible holiday names, ds,
    and countries, year combination
    """
    year_list = np.arange(1995, 2045, 1).tolist()
    all_holidays = []
    # class names in holiday packages which are not countries
    # Also cut out countries with utf-8 holidays that don't parse to ascii
    class_to_exclude = set(["rd", "BY", "BG", "JP", "RS", "UA", "KR"])

    class_list2 = inspect.getmembers(hdays_part2, inspect.isclass)
    country_set = set([name for name in list(zip(*class_list2))[0] if len(name) == 2])
    class_list1 = inspect.getmembers(hdays_part1, inspect.isclass)
    country_set1 = set([name for name in list(zip(*class_list1))[0] if len(name) == 2])
    country_set.update(country_set1)
    country_set -= class_to_exclude

    for country in country_set:
        df = make_holidays_df(year_list=year_list, country=country)
        df["country"] = country
        all_holidays.append(df)

    generated_holidays = pd.concat(all_holidays, axis=0, ignore_index=True)
    generated_holidays["year"] = generated_holidays.ds.apply(lambda x: x.year)
    generated_holidays.sort_values(["country", "ds", "holiday"], inplace=True)

    # Convert to ASCII, and drop holidays that fail to convert
    generated_holidays["holiday"] = generated_holidays["holiday"].apply(utf8_to_ascii)
    assert "FAILED_TO_PARSE" not in generated_holidays["holiday"].unique()
    generated_holidays.to_csv("../R/data-raw/generated_holidays.csv", index=False)



# encoding: utf-8

import os
from copy import copy

import httplib2
import oauth2client
import inspector

from .. import utils


def from_params(**params):
    credentials = {}
    for key, value in params.items():
        if key in (
            "client_id",
            "client_secret",
            "client_email",
            "private_key",
            "access_token",
            "refresh_token",
            "identity",
        ):
            credentials[key] = value
    return credentials


def from_environment(prefix=None, suffix=None, **params):
    keys = {
        "client_id": utils.affix(prefix, "GOOGLE_ANALYTICS_CLIENT_ID", suffix),
        "client_secret": utils.affix(prefix, "GOOGLE_ANALYTICS_CLIENT_SECRET", suffix),
        "refresh_token": utils.affix(prefix, "GOOGLE_ANALYTICS_REFRESH_TOKEN", suffix),
    }

    credentials = {}
    for credential, key in keys.items():
        value = os.environ.get(key)
        if value:
            credentials[credential] = value

    return credentials


def from_prompt(**params):
    prompted = {}

    if not params.get("identity"):
        prompted["identity"] = utils.input("Human-readable account name: ")
    if not params.get("client_id"):
        prompted["client_id"] = utils.input("Client ID: ")
    if not params.get("client_secret"):
        prompted["client_secret"] = utils.input("Client secret: ")

    return prompted


class Credentials(object):
    STRATEGIES = {
        "params": from_params,
        "environment": from_environment,
        "prompt": from_prompt,
    }

    INTERACTIVE_STRATEGIES = ["params", "environment", "prompt"]
    UNSUPERVISED_STRATEGIES = ["params", "environment"]

    @classmethod
    def find(cls, interactive=False, valid=False, complete=False, **params):
        if interactive:
            strategies = copy(cls.INTERACTIVE_STRATEGIES)
        else:
            strategies = copy(cls.UNSUPERVISED_STRATEGIES)

        attempted = ", ".join(strategies)

        credentials = cls()
        while credentials.incomplete and len(strategies):
            strategy = strategies.pop(0)
            properties = cls.STRATEGIES[strategy](**params) or {}
            for key, value in properties.items():
                if not getattr(credentials, key):
                    setattr(credentials, key, value)
                if not params.get(key):
                    params[key] = value

        # the environment variable suffix is often a good
        # descriptor of the nature of these credentials,
        # when lacking anything better
        if params.get("identity"):
            credentials.identity = params["identity"]
        elif params.get("suffix") and credentials.identity is credentials.client_id:
            credentials.identity = params.get("suffix")

        if complete and credentials.incomplete:
            raise KeyError(
                "Could not find client credentials and token. Tried {attempted}.".format(
                    attempted=attempted
                )
            )
        elif valid and credentials.invalid:
            raise KeyError(
                "Could not find client id and client secret. Tried {attempted}.".format(
                    attempted=attempted
                )
            )
        else:
            return credentials

    def __init__(
        self,
        client_id=None,
        client_secret=None,
        client_email=None,
        private_key=None,
        access_token=None,
        refresh_token=None,
        identity=None,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.client_email = client_email
        self.private_key = private_key
        self.access_token = access_token
        self.refresh_token = refresh_token
        self._identity = identity

    @property
    def token(self):
        return self.refresh_token or self.access_token

    @property
    def identity(self):
        return self._identity or self.client_id

    @identity.setter
    def identity(self, value):
        self._identity = value

    @property
    def type(self):
        if self.client_email and self.private_key:
            return 2
        elif self.client_id and self.client_secret:
            return 3
        else:
            return None

    @property
    def valid(self):
        """Valid credentials are not necessarily correct, but
        they contain all necessary information for an
        authentication attempt."""
        two_legged = self.client_email and self.private_key
        three_legged = self.client_id and self.client_secret
        return two_legged or three_legged or False

    @property
    def invalid(self):
        return not self.valid

    @property
    def complete(self):
        """ Complete credentials are valid and are either two-legged or include a token. """
        return self.valid and (
            self.access_token or self.refresh_token or self.type == 2
        )

    @property
    def incomplete(self):
        return not self.complete

    @property
    def oauth(self):
        if self.incomplete:
            return None
        else:
            if self.type == 2:
                return oauth2client.client.SignedJwtAssertionCredentials(
                    service_account_name=self.client_email,
                    private_key=self.private_key.encode("utf-8"),
                    scope="https://www.googleapis.com/auth/analytics.readonly",
                )
            else:
                return oauth2client.client.OAuth2Credentials(
                    access_token=self.access_token,
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                    refresh_token=self.refresh_token,
                    token_expiry=None,
                    token_uri=oauth2client.GOOGLE_TOKEN_URI,
                    user_agent=None,
                    revoke_uri=oauth2client.GOOGLE_REVOKE_URI,
                    id_token=None,
                    token_response=None,
                )

    def serialize(self):
        return {
            "identity": self.identity,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "client_email": self.client_email,
            "private_key": self.private_key,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
        }

    def authorize(self):
        return self.oauth.authorize(httplib2.Http())

    def revoke(self):
        if not self.token:
            raise KeyError("Cannot revoke a token when no token was provided.")

        # `credentials.revoke` will try to revoke the refresh token even
        # if it's None, which will fail, so we have to miss with the innards
        # of oauth2client here a little bit
        return self.oauth._do_revoke(httplib2.Http().request, self.token)


def normalize(fn):
    @inspector.changes(fn)
    def normalized_fn(
        client_id=None,
        client_secret=None,
        access_token=None,
        refresh_token=None,
        identity=None,
    ):

        if isinstance(client_id, Credentials):
            credentials = client_id
        else:
            credentials = Credentials(
                client_id, client_secret, access_token, refresh_token, identity
            )

        return fn(credentials)

    return normalized_fn



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: NBEATS Model"""

import os
import numpy as np

import torch
from torch import optim
from torch.nn import functional as F
from tqdm.auto import tqdm
from nbeats_pytorch.model import (
    NBeatsNet,
)


from forecastga.models.base import BaseModel

from forecastga.helpers.data import data_generator


CHECKPOINT_NAME = "nbeats-training-checkpoint.th"


class NBEATS_Model(BaseModel):
    """NBEATS Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        """Train NBEATS Model"""

        if os.path.isfile(CHECKPOINT_NAME):
            os.remove(CHECKPOINT_NAME)

        steps = kwargs.get("steps", 50)
        batch_size = kwargs.get("batch_size", 10)
        patience = kwargs.get("patience", 5)
        device = self.get_device()

        # There seems to be an issue with
        net = NBeatsNet(
            stack_types=[
                NBeatsNet.TREND_BLOCK,
                NBeatsNet.SEASONALITY_BLOCK,
                NBeatsNet.GENERIC_BLOCK,
            ],
            forecast_length=self.forecast_len,
            thetas_dims=kwargs.get("thetas_dims", [2, 8, 3]),
            nb_blocks_per_stack=kwargs.get("nb_blocks_per_stack", 3),
            backcast_length=self.forecast_len,
            hidden_layer_units=kwargs.get("hidden_layer_units", 128),
            share_weights_in_stack=False,
            device=device,
        )

        x_batch, y_batch, norm_constant = self.format_input(
            self.dataframe, self.forecast_len
        )

        c = len(x_batch)
        if self.in_sample:
            c -= self.forecast_len

        optimiser = optim.Adam(net.parameters())

        data = data_generator(x_batch[:c], y_batch[:c], batch_size)

        best_loss = float("inf")
        counter = 0

        for _ in tqdm(range(steps)):
            loss = self.train_100_grad_steps(data, device, net, optimiser)
            if loss < best_loss:
                best_loss, counter = loss, 0
            else:
                counter += 1
                if counter >= patience:
                    break

        self.model = net
        self.constant = norm_constant

    def forecast(self, **kwargs):
        """Forecast NEATS Model"""

        x_batch, _, _ = self.format_input(
            self.dataframe, self.forecast_len, constant=self.constant
        )

        c = len(x_batch)
        if self.in_sample:
            c -= self.forecast_len

        self.model.eval()

        if self.in_sample:
            _, forecast = self.model(torch.tensor(x_batch[c:], dtype=torch.float))
        else:
            _, forecast = self.model(torch.tensor(x_batch, dtype=torch.float))

        p = forecast.cpu().detach().numpy() if self.GPU else forecast.detach().numpy()
        self.prediction = p[-1] * self.constant

    def get_device(self):
        return torch.device("cuda") if self.GPU else torch.device("cpu")

    @staticmethod
    def format_input(df, forecast_length, constant=None):

        backcast_length = 1 * forecast_length

        x = df.values
        norm_constant = constant if constant else np.max(x)
        x = x / norm_constant

        x_batch, y_batch = [], []

        # Batches the results into x_train_batch: x and y_train_batch: x + forecast_length
        for i in range(backcast_length + 1, len(x) - forecast_length + 1):
            x_batch.append(x[i - backcast_length : i])
            y_batch.append(x[i : i + forecast_length])

        x_batch = np.array(x_batch)[..., 0]
        y_batch = np.array(y_batch)[..., 0]

        return x_batch, y_batch, norm_constant

    def train_100_grad_steps(self, data, device, net, optimiser):
        """Trainer"""
        global_step = self.load(net, optimiser)
        global_step_init = global_step
        step_loss = 0
        for x_train_batch, y_train_batch in data:
            global_step += 1
            optimiser.zero_grad()
            net.train()
            _, forecast = net(torch.tensor(x_train_batch, dtype=torch.float).to(device))
            loss = F.mse_loss(
                forecast, torch.tensor(y_train_batch, dtype=torch.float).to(device)
            )
            step_loss += loss.item()
            loss.backward()
            optimiser.step()
            if global_step > 0 and global_step % 100 == 0:
                with torch.no_grad():
                    self.save(net, optimiser, global_step)
                break

        return step_loss / (global_step - global_step_init)

    def load(self, model, optimiser):
        """loader/saver for checkpoints"""

        if os.path.exists(CHECKPOINT_NAME):
            checkpoint = torch.load(CHECKPOINT_NAME)
            model.load_state_dict(checkpoint["model_state_dict"])
            optimiser.load_state_dict(checkpoint["optimizer_state_dict"])
            grad_step = checkpoint["grad_step"]

            return grad_step
        return 0

    def save(self, model, optimiser, grad_step):
        torch.save(
            {
                "grad_step": grad_step,
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimiser.state_dict(),
            },
            CHECKPOINT_NAME,
        )



# encoding: utf-8

import functools
import re

import addressable


from snakify import snakify

from . import utils

TYPES = {
    "STRING": utils.unicode,
    "INTEGER": int,
    "FLOAT": float,
    "PERCENT": float,
    "TIME": float,
    "CURRENCY": float,
}

DIMENSIONS = {
    "ga:date": lambda date: utils.date.parse(date).date(),
    "ga:dateHour": lambda date: utils.date.parse("{} {}".format(date[:8], date[8:])),
}


def escape_chars(value, chars=",;"):
    if value is True:
        return "Yes"
    if value is False:
        return "No"

    value = utils.unicode(value)
    for char in chars:
        value = value.replace(char, "\\" + char)

    return value


def escape(method):
    @functools.wraps(method)
    def escaped_method(self, *values):
        values = utils.builtins.map(escape_chars, values)
        return method(self, *values)

    return escaped_method


class Column:
    selectors = (
        "eq",
        "neq",
        "lt",
        "lte",
        "gt",
        "gte",
        "between",
        "any",
        "contains",
        "ncontains",
        "re",
        "nre",
    )

    @classmethod
    def from_metadata(cls, metadata):
        attributes = metadata["attributes"]
        data_format = (
            DIMENSIONS.get(metadata["id"])
            or TYPES.get(attributes["dataType"])
            or utils.identity
        )
        is_deprecated = attributes.get("status", "ACTIVE") == "DEPRECATED"
        is_allowed_in_segments = "allowedInSegments" in attributes
        column = Column(
            metadata["id"],
            column_type=attributes["type"].lower(),
            format=data_format,
            attributes=attributes,
            deprecated=is_deprecated,
            allowed_in_segments=is_allowed_in_segments,
        )
        return column.expand()

    def __init__(
        self,
        column_id,
        column_type,
        format=utils.unicode,
        attributes={},
        deprecated=False,
        allowed_in_segments=True,
    ):
        self.account = None
        self.id = column_id
        self.report_type, self.slug = self.id.split(":")
        index = re.search(r"\d{1,2}", self.slug)
        if index:
            self.index = int(index.group(0))
        else:
            self.index = None
        self.python_slug = snakify(self.slug)
        self.attributes = attributes
        self.name = attributes.get("uiName", column_id).replace("XX", str(self.index))
        self.group = attributes.get("group")
        self.description = attributes.get("description")
        self.type = column_type
        # TODO: evaluate if we can improve casting
        self.cast = format
        self.is_deprecated = deprecated
        self.is_allowed_in_segments = allowed_in_segments

    def link(self, account):
        self.account = account

    def expand(self):
        columns = []
        if "XX" in self.id:
            min_index = int(self.attributes.get("minTemplateIndex", "1"))
            max_index = int(self.attributes.get("maxTemplateIndex", "20"))
            for i in range(min_index, max_index + 1):
                column = Column(
                    self.id.replace("XX", str(i)),
                    column_type=self.type,
                    format=self.cast,
                    attributes=self.attributes,
                    deprecated=self.is_deprecated,
                    allowed_in_segments=self.is_allowed_in_segments,
                )
                columns.append(column)
        else:
            columns = [self]

        return columns

    @escape
    def eq(self, value):
        return "{id}=={value}".format(id=self.id, value=value)

    @escape
    def neq(self, value):
        return "{id}!={value}".format(id=self.id, value=value)

    @escape
    def lt(self, value):
        return "{id}<{value}".format(id=self.id, value=value)

    @escape
    def lte(self, value):
        return "{id}<={value}".format(id=self.id, value=value)

    @escape
    def gt(self, value):
        return "{id}>{value}".format(id=self.id, value=value)

    @escape
    def gte(self, value):
        return "{id}>={value}".format(id=self.id, value=value)

    @escape
    def between(self, a, b):
        return "{id}<>{a}_{b}".format(id=self.id, a=a, b=b)

    @escape
    def any(self, *values):
        return "{id}[]{values}".format(id=self.id, values="|".join(values))

    @escape
    def contains(self, value):
        return "{id}=@{value}".format(id=self.id, value=value)

    @escape
    def ncontains(self, value):
        return "{id}!@{value}".format(id=self.id, value=value)

    @escape
    def re(self, value):
        return "{id}=~{value}".format(id=self.id, value=value)

    @escape
    def nre(self, value):
        return "{id}!~{value}".format(id=self.id, value=value)

    # useful when sorting a query
    def __neg__(self):
        return "-" + self.id

    def __repr__(self):
        report_types = {
            "ga": "Core",
            "rt": "Realtime",
            None: "Unbound",
        }
        return "<googleanalytics.columns.{query_type} object: {column_type}, {name} ({id})>".format(
            query_type=report_types[self.report_type],
            column_type=self.type.capitalize(),
            name=self.name,
            id=self.id,
        )


# see https://developers.google.com/analytics/devguides/reporting/core/v3/segments#reference
class Segment(Column):
    # CHECK: do we need to call super here?
    def __init__(self, raw, account):
        self.raw = raw
        self.id = raw["segmentId"]
        self.report_type, self.slug = self.id.split("::")
        self.python_slug = snakify(self.slug)
        self.name = raw["name"]
        self.kind = raw["kind"].lower()
        self.definition = raw["definition"]

    def __repr__(self):
        return "<googleanalytics.columns.Segment object: {name} ({id})>".format(
            **self.__dict__
        )


class Filter:
    pass


class Goal:
    pass

    """
     goals = service.management().goals().list(
            accountId=firstAccountId,
            webPropertyId=firstWebpropertyId,
            profileId=firstProfileId).execute()
    """


class ColumnList(addressable.List):
    COLUMN_TYPE = Column

    def __init__(self, columns, **options):
        options["items"] = columns
        options["name"] = self.COLUMN_TYPE.__class__.__name__
        options["indices"] = ("name", "id", "slug", "python_slug")
        options["insensitive"] = True
        super(ColumnList, self).__init__(**options)

    @utils.vectorize
    def normalize(self, value):
        if isinstance(value, self.COLUMN_TYPE):
            return value

        return self[value]

    @utils.vectorize
    def serialize(self, value, greedy=True):
        """
        Greedy serialization requires the value to either be a column
        or convertible to a column, whereas non-greedy serialization
        will pass through any string as-is and will only serialize
        Column objects.

        Non-greedy serialization is useful when preparing queries with
        custom filters or segments.
        """

        if greedy and not isinstance(value, Column):
            value = self.normalize(value)

        if isinstance(value, Column):
            return value.id

        return value


class SegmentList(ColumnList):
    COLUMN_TYPE = Segment


def is_deprecated(column):
    return column.is_deprecated


def is_supported(column):
    return not column.is_deprecated


def is_metric(column):
    return column.type == "metric"


def is_dimension(column):
    return column.type == "dimension"


def is_core(column):
    return column.report_type == "ga"


def is_live(column):
    return column.report_type == "rt"



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: SSA Helper"""

import numpy as np
import pandas as pd
from numpy import matrix as m
from scipy import linalg


class mySSA(object):
    """Singular Spectrum Analysis object"""

    def __init__(self, time_series):

        self.ts = pd.DataFrame(time_series)
        self.ts_name = self.ts.columns.tolist()[0]
        if self.ts_name == 0:
            self.ts_name = "ts"
        self.ts_v = self.ts.values
        self.ts_N = self.ts.shape[0]
        self.freq = self.ts.index.inferred_freq

    @staticmethod
    def _dot(x, y):
        """Alternative formulation of dot product to allow missing values in arrays/matrices"""
        pass

    @staticmethod
    def get_contributions(X=None, s=None, plot=True):
        """Calculate the relative contribution of each of the singular values"""
        lambdas = np.power(s, 2)
        frob_norm = np.linalg.norm(X)
        ret = pd.DataFrame(lambdas / (frob_norm ** 2), columns=["Contribution"])
        ret["Contribution"] = ret.Contribution.round(4)
        if plot:
            ax = ret[ret.Contribution != 0].plot.bar(legend=False)
            ax.set_xlabel("Lambda_i")
            ax.set_title("Non-zero contributions of Lambda_i")
            vals = ax.get_yticks()
            ax.set_yticklabels(["{:3.2f}%".format(x * 100) for x in vals])
            return ax
        return ret[ret.Contribution > 0]

    @staticmethod
    def diagonal_averaging(hankel_matrix):
        """Performs anti-diagonal averaging from given hankel matrix
        Returns: Pandas DataFrame object containing the reconstructed series"""
        mat = m(hankel_matrix)
        L, K = mat.shape
        L_star, K_star = min(L, K), max(L, K)
        # new = np.zeros((L, K))
        if L > K:
            mat = mat.T
        ret = []

        # Diagonal Averaging
        for k in range(1 - K_star, L_star):
            mask = np.eye(K_star, k=k, dtype="bool")[::-1][:L_star, :]
            mask_n = sum(sum(mask))
            ma = np.ma.masked_array(mat.A, mask=1 - mask)
            ret += [ma.sum() / mask_n]

        return pd.DataFrame(ret).rename(columns={0: "Reconstruction"})

    def view_time_series(self):
        """Plot the time series"""
        self.ts.plot(title="Original Time Series")

    def embed(
        self,
        embedding_dimension=None,
        suspected_frequency=None,
        verbose=False,
        return_df=False,
    ):
        """Embed the time series with embedding_dimension window size.
        Optional: suspected_frequency changes embedding_dimension such that it is divisible by suspected frequency"""
        if not embedding_dimension:
            self.embedding_dimension = self.ts_N // 2
        else:
            self.embedding_dimension = embedding_dimension
        if suspected_frequency:
            self.suspected_frequency = suspected_frequency
            self.embedding_dimension = (
                self.embedding_dimension // self.suspected_frequency
            ) * self.suspected_frequency

        self.K = self.ts_N - self.embedding_dimension + 1
        self.X = m(linalg.hankel(self.ts, np.zeros(self.embedding_dimension))).T[
            :, : self.K
        ]
        self.X_df = pd.DataFrame(self.X)
        self.X_complete = self.X_df.dropna(axis=1)
        self.X_com = m(self.X_complete.values)
        self.X_missing = self.X_df.drop(self.X_complete.columns, axis=1)
        self.X_miss = m(self.X_missing.values)
        self.trajectory_dimentions = self.X_df.shape
        self.complete_dimensions = self.X_complete.shape
        self.missing_dimensions = self.X_missing.shape
        self.no_missing = self.missing_dimensions[1] == 0

        if return_df:
            return self.X_df

    def decompose(self, verbose=False):
        """Perform the Singular Value Decomposition and identify the rank of the embedding subspace
        Characteristic of projection: the proportion of variance captured in the subspace"""
        X = self.X_com
        self.S = X * X.T
        self.U, self.s, self.V = linalg.svd(self.S)
        self.U, self.s, self.V = m(self.U), np.sqrt(self.s), m(self.V)
        self.d = np.linalg.matrix_rank(X)
        Vs, Xs, Ys, Zs = {}, {}, {}, {}
        for i in range(self.d):
            Zs[i] = self.s[i] * self.V[:, i]
            Vs[i] = X.T * (self.U[:, i] / self.s[i])
            Ys[i] = self.s[i] * self.U[:, i]
            Xs[i] = Ys[i] * (m(Vs[i]).T)
        self.Vs, self.Xs = Vs, Xs
        self.s_contributions = self.get_contributions(X, self.s, False)
        self.r = len(self.s_contributions[self.s_contributions > 0])
        self.r_characteristic = round(
            (self.s[: self.r] ** 2).sum() / (self.s ** 2).sum(), 4
        )
        self.orthonormal_base = {i: self.U[:, i] for i in range(self.r)}

    def view_s_contributions(
        self, adjust_scale=False, cumulative=False, return_df=False
    ):
        """View the contribution to variance of each singular value and its corresponding signal"""
        contribs = self.s_contributions.copy()
        contribs = contribs[contribs.Contribution != 0]
        if cumulative:
            contribs["Contribution"] = contribs.Contribution.cumsum()
        if adjust_scale:
            contribs = (1 / contribs).max() * 1.1 - (1 / contribs)
        ax = contribs.plot.bar(legend=False)
        ax.set_xlabel("Singular_i")
        ax.set_title(
            "Non-zero{} contribution of Singular_i {}".format(
                " cumulative" if cumulative else "", "(scaled)" if adjust_scale else ""
            )
        )
        if adjust_scale:
            ax.axes.get_yaxis().set_visible(False)
        vals = ax.get_yticks()
        ax.set_yticklabels(["{:3.0f}%".format(x * 100) for x in vals])
        if return_df:
            return contribs

    @classmethod
    def view_reconstruction(
        cls, *hankel, names=None, return_df=False, plot=True, symmetric_plots=False
    ):
        """Visualise the reconstruction of the hankel matrix/matrices passed to *hankel"""
        hankel_mat = None
        for han in hankel:
            if isinstance(hankel_mat, m):
                hankel_mat = hankel_mat + han
            else:
                hankel_mat = han.copy()
        hankel_full = cls.diagonal_averaging(hankel_mat)
        title = "Reconstruction of signal"
        if names or names == 0:
            title += " associated with singular value{}: {}"
            title = title.format("" if len(str(names)) == 1 else "s", names)
        if plot:
            ax = hankel_full.plot(legend=False, title=title)
            if symmetric_plots:
                velocity = hankel_full.abs().max()[0]
                ax.set_ylim(bottom=-velocity, top=velocity)
        if return_df:
            return hankel_full

    def _forecast_prep(self, singular_values=None):
        self.X_com_hat = np.zeros(self.complete_dimensions)
        self.verticality_coefficient = 0
        self.forecast_orthonormal_base = {}
        if singular_values:
            try:
                for i in singular_values:
                    self.forecast_orthonormal_base[i] = self.orthonormal_base[i]
            except:
                if singular_values == 0:
                    self.forecast_orthonormal_base[0] = self.orthonormal_base[0]
                else:
                    raise (
                        "Please pass in a list/array of singular value indices to use for forecast"
                    )
        else:
            self.forecast_orthonormal_base = self.orthonormal_base
        self.R = np.zeros(self.forecast_orthonormal_base[0].shape)[:-1]
        for Pi in self.forecast_orthonormal_base.values():
            self.X_com_hat += Pi * Pi.T * self.X_com
            pi = np.ravel(Pi)[-1]
            self.verticality_coefficient += pi ** 2
            self.R += pi * Pi[:-1]
        self.R = m(self.R / (1 - self.verticality_coefficient))
        self.X_com_tilde = self.diagonal_averaging(self.X_com_hat)

    def forecast_recurrent(
        self,
        steps_ahead=12,
        singular_values=None,
        plot=False,
        return_df=False,
        **plotargs
    ):
        """Forecast from last point of original time series up to steps_ahead using recurrent methodology
        This method also fills any missing data from the original time series."""
        try:
            self.X_com_hat
        except (AttributeError):
            self._forecast_prep(singular_values)
        self.ts_forecast = np.array(self.ts_v[0])
        for i in range(1, self.ts_N + steps_ahead):
            try:
                if np.isnan(self.ts_v[i]):
                    x = (
                        self.R.T
                        * m(self.ts_forecast[max(0, i - self.R.shape[0]) : i]).T
                    )
                    self.ts_forecast = np.append(self.ts_forecast, x[0])
                else:
                    self.ts_forecast = np.append(self.ts_forecast, self.ts_v[i])
            except (IndexError):
                x = self.R.T * m(self.ts_forecast[i - self.R.shape[0] : i]).T
                self.ts_forecast = np.append(self.ts_forecast, x[0])
        self.forecast_N = i + 1
        new_index = pd.date_range(
            start=self.ts.index.min(), periods=self.forecast_N, freq=self.freq
        )
        forecast_df = pd.DataFrame(
            self.ts_forecast, columns=["Forecast"], index=new_index
        )
        forecast_df["Original"] = np.append(self.ts_v, [np.nan] * steps_ahead)
        if plot:
            forecast_df.plot(title="Forecasted vs. original time series", **plotargs)
        if return_df:
            return forecast_df



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: SARIMAX Model"""

from forecastga.models.base import BaseModel


class SARIMAX_Model:
    """ARIMA Model Class"""

    def __init__(self):
        raise NotImplementedError

    def dataframe(self):
        raise NotImplementedError

    def train(self):
        raise NotImplementedError

    def forecast(self):
        raise NotImplementedError



# encoding: utf-8


def single_serve(message=None, port=5000):
    import logging
    from werkzeug.wrappers import Request, Response
    from werkzeug.serving import run_simple

    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)

    captured = {}

    def application(environ, start_response):
        request = Request(environ)
        request.environ.get("werkzeug.server.shutdown")()
        captured.update(dict(request.args.items()))
        if message:
            print(message)
        response = Response(message, mimetype="text/plain")
        return response(environ, start_response)

    run_simple("localhost", port, application)
    return captured



# encoding: utf-8

"""
Chaining for filters and segments.
"""


def condition(value):
    return "condition::" + value


def sequence(value):
    return "sequence::" + value


def all(*values):
    return condition(";".join(values))


def any(*values):
    return condition(",".join(values))


def followed_by(*values):
    return sequence(";->>".join(values))


def immediately_followed_by(*values):
    return sequence(";->".join(values))



# encoding: utf-8
""" Google Analytics Package"""

from . import (
    auth,
    utils,
    account,
    blueprint,
    columns,
    errors,
    query,
    segments,
)
from .auth import authenticate, authorize, revoke
from .blueprint import Blueprint



# encoding: utf-8

"""Google Analytics Account"""

import yaml
import addressable

from . import utils
from . import query
from . import columns
from .columns import Column, Segment, ColumnList, SegmentList


class Account:
    """
    An account is usually but not always associated with a single
    website. It will often contain multiple web properties
    (different parts of your website that you've configured
    Google Analytics to analyze separately, or simply the default
    web property that every website has in Google Analytics),
    which in turn will have one or more profiles.

    You should navigate to a profile to run queries.

    ```python
    import googleanalytics as ga
    accounts = ga.authenticate()
    profile = accounts['debrouwere.org'].webproperties['UA-12933299-1'].profiles['debrouwere.org']
    report = profile.core.query('pageviews').range('2014-10-01', '2014-10-31').get()
    print(report['pageviews'])
    ```
    """

    def __init__(self, raw, service, credentials):
        self.service = service
        self.credentials = credentials
        self.raw = raw
        self.id = raw["id"]
        self.name = raw["name"]
        self.permissions = raw["permissions"]["effective"]

    @property
    @utils.memoize
    def webproperties(self):
        """
        A list of all web properties on this account. You may
        select a specific web property using its name, its id
        or an index.

        ```python
        account.webproperties[0]
        account.webproperties['UA-9234823-5']
        account.webproperties['debrouwere.org']
        ```
        """

        raw_properties = (
            self.service.management()
            .webproperties()
            .list(accountId=self.id)
            .execute()["items"]
        )
        _webproperties = [WebProperty(raw, self) for raw in raw_properties]
        return addressable.List(
            _webproperties, indices=["id", "name"], insensitive=True
        )

    @property
    def query(self, *vargs, **kwargs):
        """ A shortcut to the first profile of the first webproperty. """
        return self.webproperties[0].query(*vargs, **kwargs)

    def __repr__(self):
        return "<googleanalytics.account.Account object: {} ({})>".format(
            self.name, self.id
        )


class WebProperty:
    """
    A web property is a particular website you're tracking in Google Analytics.
    It has one or more profiles, and you will need to pick one from which to
    launch your queries.
    """

    def __init__(self, raw, account):
        self.account = account
        self.raw = raw
        self.id = raw["id"]
        self.name = raw["name"]
        # on rare occassions, e.g. for abandoned web properties,
        # a website url might not be present
        self.url = raw.get("websiteUrl")

    @property
    def profile(self):
        default = self.raw["defaultProfileId"]
        return self.profiles[default]

    @property
    @utils.memoize
    def profiles(self):
        """
        A list of all profiles on this web property. You may
        select a specific profile using its name, its id
        or an index.

        ```python
        property.profiles[0]
        property.profiles['9234823']
        property.profiles['marketing profile']
        ```
        """
        raw_profiles = (
            self.account.service.management()
            .profiles()
            .list(accountId=self.account.id, webPropertyId=self.id)
            .execute()["items"]
        )
        profiles = [Profile(raw, self) for raw in raw_profiles]
        return addressable.List(profiles, indices=["id", "name"], insensitive=True)

    def query(self, *vargs, **kwargs):
        """
        A shortcut to the first profile of this webproperty.
        """
        return self.profiles[0].query(*vargs, **kwargs)

    def __repr__(self):
        return "<googleanalytics.account.WebProperty object: {} ({})>".format(
            self.name, self.id
        )


class Profile:
    """
    A profile is a particular analytics configuration of a web property.
    Each profile belongs to a web property and an account. As all
    queries using the Google Analytics API run against a particular
    profile, queries can only be created from a `Profile` object.

    ```python
    profile.query('pageviews').range('2014-01-01', days=7).get()
    ```
    """

    def __init__(self, raw, webproperty):
        self.raw = raw
        self.webproperty = webproperty
        self.account = webproperty.account
        self.id = raw["id"]
        self.name = raw["name"]
        self.core = CoreReportingAPI(self)
        self.realtime = RealTimeReportingAPI(self)

    def __repr__(self):
        return "<googleanalytics.account.Profile object: {} ({})>".format(
            self.name, self.id
        )


class ReportingAPI:
    REPORT_TYPES = {
        "ga": "ga",
        "realtime": "rt",
    }

    QUERY_TYPES = {
        "ga": query.CoreQuery,
        "realtime": query.RealTimeQuery,
    }

    def __init__(self, endpoint, profile):
        """
        Endpoint can be one of `ga` or `realtime`.
        """

        # various shortcuts
        self.profile = profile
        self.account = profile.account
        self.service = service = profile.account.service
        root = service.data()
        self.endpoint_type = endpoint
        self.endpoint = getattr(root, endpoint)()

        # query interface
        self.report_type = self.REPORT_TYPES[endpoint]
        query_class = self.QUERY_TYPES[endpoint]
        self.query = query_class(self)

        # optional caching layer
        self.cache = None

    @property
    @utils.memoize
    def all_columns(self):
        query_service = (
            self.service.metadata().columns().list(reportType=self.report_type)
        )
        raw_columns = query_service.execute()["items"]
        hydrated_columns = utils.flatten(map(Column.from_metadata, raw_columns))
        return ColumnList(hydrated_columns, unique=False)

    @property
    @utils.memoize
    def columns(self):
        return addressable.filter(columns.is_supported, self.all_columns)

    @property
    @utils.memoize
    def segments(self):
        query_service = self.service.management().segments().list()
        raw_segments = query_service.execute()["items"]
        hydrated_segments = [Segment(raw, self) for raw in raw_segments]
        return SegmentList(hydrated_segments)

    @property
    @utils.memoize
    def metrics(self):
        return addressable.filter(columns.is_metric, self.columns)

    @property
    @utils.memoize
    def dimensions(self):
        return addressable.filter(columns.is_dimension, self.columns)

    @property
    @utils.memoize
    def goals(self):
        raise NotImplementedError()

    def __repr__(self):
        return "<googleanalytics.account.{} object>".format(self.__class__.__name__)


class CoreReportingAPI(ReportingAPI):
    def __init__(self, profile):
        super(CoreReportingAPI, self).__init__("ga", profile)


class RealTimeReportingAPI(ReportingAPI):
    def __init__(self, profile):
        super(RealTimeReportingAPI, self).__init__("realtime", profile)

    # in principle, we should be able to reuse everything from the ReportingAPI
    # base class, but the Real Time Reporting API is still in beta and some
    # things – like a metadata endpoint – are missing.
    @property
    @utils.memoize
    def all_columns(self):
        raw_columns = yaml.load(open(utils.here("realtime.yml")))
        hydrated_columns = utils.flatten(map(Column.from_metadata, raw_columns))
        return ColumnList(hydrated_columns)



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: Gluonts Model"""
import pandas as pd

from gluonts.model.deepar import DeepAREstimator
from gluonts.trainer import Trainer
from gluonts.dataset.common import ListDataset

from forecastga.models.base import BaseModel


class Gluonts_Model(BaseModel):
    """Gluonts Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):

        epochs = kwargs.get("epochs", 10)

        # Adjust class freq.
        self.freq = pd.infer_freq(self.train_df.index)
        if self.freq == "MS":
            self.freq = "M"

        estimator = DeepAREstimator(
            freq=self.freq,
            prediction_length=self.forecast_len,
            trainer=Trainer(
                epochs=epochs, batch_size=64, ctx="gpu" if self.GPU else "cpu"
            ),
        )

        self.model = estimator.train(
            training_data=self.format_input(self.train_df, self.freq)
        )

    def forecast(self):

        if self.in_sample:
            forecast = self.model.predict(
                self.format_input(
                    self.dataframe["Target"],
                    self.freq,
                    self.train_df.index[-1]
                    + self.train_df.index.to_series().diff().min(),
                )
            )
        else:
            forecast = self.model.predict(
                self.format_input(
                    self.dataframe["Target"],
                    self.freq,
                    self.dataframe["Target"].index[-1]
                    + self.train_df.index.to_series().diff().min(),
                )
            )

        self.prediction = list(forecast)[0].samples.mean(axis=0)  # .quantile(0.5)

    @staticmethod
    def format_input(df, freq, target=None):
        if target:
            return ListDataset(
                [{"start": df.index[0], "target": df.to_frame().Target[:target]}],
                freq=freq,
            )

        return ListDataset(
            [{"start": df.index[0], "target": df.to_frame().Target}], freq=freq
        )



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: ARIMA Model"""

import pmdarima as pm

from forecastga.models.base import BaseModel


class ARIMA_Model(BaseModel):
    """ARIMA Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        self.model = pm.auto_arima(self.train_df, seasonal=True, m=self.seasons)

    def forecast(self):
        self.prediction = self.model.predict(self.forecast_len)



#! /usr/bin/env python
# coding: utf-8
#

"""TForecastGA: BATS2 Model (TBATS1 With Two Seasonal Periods)"""

from tbats import TBATS

from forecastga.helpers.data import get_unique_N, season_list

from forecastga.models.base import BaseModel


class TBATS2_Model(BaseModel):
    """TBATS2 Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        bat = TBATS(
            seasonal_periods=list(get_unique_N(season_list(self.train_df), 2)),
            use_arma_errors=False,
            use_box_cox=True,
            use_trend=True,
        )
        self.model = bat.fit(self.train_df)

    def forecast(self):
        self.prediction = self.model.forecast(self.forecast_len)



import textwrap


# Python 2 and 3 compatibility
try:
    unicode = unicode
except NameError:
    unicode = str


def format(string, **kwargs):
    return textwrap.dedent(string).format(**kwargs).strip()


def affix(prefix, base, suffix, connector="_"):
    if prefix:
        prefix = prefix + connector
    else:
        prefix = ""

    if suffix:
        suffix = connector + suffix
    else:
        suffix = ""

    return prefix + base + suffix


# a supercharged `join` function, analogous to `paste` in the R language
def paste(rows, *delimiters):
    delimiter = delimiters[-1]
    delimiters = delimiters[:-1]

    if len(delimiters):
        return paste([paste(row, *delimiters) for row in rows], delimiter)
    else:
        return delimiter.join(map(unicode, rows))


# a supercharged `split` function, the inverse of `paste`
def cut(s, *delimiters):
    delimiter = delimiters[-1]
    delimiters = delimiters[:-1]

    if len(delimiters):
        return [cut(ss, *delimiters) for ss in cut(s, delimiter)]
    else:
        return s.split(delimiter)



# encoding: utf-8

import json
import webbrowser

import addressable
from oauth2client import client
from apiclient import discovery

from forecastga.ga import utils, account
from .credentials import Credentials, normalize


class Flow(client.OAuth2WebServerFlow):
    def __init__(self, client_id, client_secret, redirect_uri):
        super(Flow, self).__init__(
            client_id,
            client_secret,
            scope="https://www.googleapis.com/auth/analytics.readonly",
            redirect_uri=redirect_uri,
        )

    def step2_exchange(self, code):
        credentials = super(Flow, self).step2_exchange(code)
        return Credentials.find(complete=True, **credentials.__dict__)


# a simplified version of `oauth2client.tools.run_flow`
def authorize(client_id, client_secret):
    flow = Flow(client_id, client_secret, redirect_uri="urn:ietf:wg:oauth:2.0:oob")

    authorize_url = flow.step1_get_authorize_url()
    print("Go to the following link in your browser: " + authorize_url)
    code = input("Enter verification code: ").strip()
    return flow.step2_exchange(code)


@normalize
def revoke(credentials):
    return credentials.revoke()


@normalize
def authenticate(credentials):
    client = credentials.authorize()
    service = discovery.build("analytics", "v3", http=client, cache_discovery=False)
    raw_accounts = service.management().accounts().list().execute()["items"]
    accounts = [account.Account(raw, service, credentials) for raw in raw_accounts]
    return addressable.List(accounts, indices=["id", "name"], insensitive=True)



#! /usr/bin/env python
# coding: utf-8
#

"""ARIMA Model"""

from forecastga.models.base import BaseModel


class ARIMA_Model:
    """ARIMA Model Class"""

    def __init__(self, config):
        super().__init__(config)

        """
        Available model attributes:

        self.seasonality (str)
        self.forecast_len (int)
        self.freq (str)
        self.train_proportion (float)
        self.in_sample (bool)
        self.GPU (bool)

        self.dataframe (pd.Series)
        self.train_df (pd.Series)
        self.forecast_df (pd.Series) or None
        self.seasons (int)
        self.periods (int)
        """

    def train(self):
        self.model = pm.auto_arima(self.train_df, seasonal=True, m=self.seasons)

    def forecast(self):
        if self.insample:
            self.prediction = self.model.predict(self.forecast_len)
            # Prediction can be a list, np.Array, or pandas series.
        else:
            # Do something else if outsample.
            pass

    @staticmethod
    def format_input(df, forecast_length, constant=None):
        pass

    @staticmethod
    def format_output(df, forecast_length, constant=None):
        pass



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: PYAF Model"""

import pyaf.ForecastEngine as autof

from forecastga.models.base import BaseModel


class PYAF_Model(BaseModel):
    """PYAF Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        self.model = autof()
        self.model.train(
            iInputDS=self.train_df.reset_index(),
            iTime="Date",
            iSignal="Target",
            iHorizon=len(self.train_df),
        )

    def forecast(self):
        self.model.forecast(
            iInputDS=self.train_df.reset_index(), iHorizon=self.forecast_len
        )
        self.prediction = self.model["Target_Forecast"][-self.forecast_len :].values



#! /usr/bin/env python
# coding: utf-8
#
"""ForecastGA: Ensembles"""

import pandas as pd
from sklearn.model_selection import train_test_split as tts
from sklearn.decomposition import PCA
from statsmodels.tools.eval_measures import rmse
from tsfresh.utilities.dataframe_functions import impute, roll_time_series
from tsfresh import extract_features
from tsfresh import select_features
import lightgbm as lgb

from forecastga.helpers.logging import get_logger
from forecastga.helpers.data import constant_feature_detect

_LOG = get_logger(__name__)


def ensemble_performance(forecasts):
    dict_perf = {}
    for col, _ in forecasts.iteritems():
        dict_perf[col] = {}
        dict_perf[col]["rmse"] = rmse(forecasts["Target"], forecasts[col])
        dict_perf[col]["mse"] = dict_perf[col]["rmse"] ** 2
        dict_perf[col]["mean"] = forecasts[col].mean()
    return pd.DataFrame.from_dict(dict_perf)


def time_feature(df, perd):
    if perd in ["MS", "M", "BM", "BMS"]:
        df["month"] = df.index.month
    elif perd in ["BH", "H"]:
        df["hour"] = df.index.hour
    elif perd == "B":
        df["dayofweek"] = df.index.dayofweek
    elif perd == "D":
        df["dayofweek"] = df.index.dayofweek
    elif perd in ["W", "W-SUN", "W-MON", "W-TUE", "W-WED", "W-THU", "W-FRI", "W-SAT"]:
        df["week"] = df.index.week
    elif perd in ["Q", "QS", "BQ", "BQS"]:
        df["quarter"] = df.index.quarter
    elif perd in ["T", "min"]:
        df["minute"] = df.index.minute
    elif perd == "S":
        df["second"] = df.index.second
    # elif perd in ["L","ms"]:
    #   periodocity = 1000
    # elif perd in ["U","us"]:
    #   periodocity = 1000
    # elif perd=="N":
    #   periodocity = 1000
    return df


def ensemble_lightgbm(forecast_in, forecast_out, pred):

    forecast_in_copy = forecast_in.copy()

    forecast_in_copy = time_feature(forecast_in_copy, pred)
    forecast_in_copy["mean"] = forecast_in_copy.drop(["Target"], axis=1).mean(axis=1)
    forecast_train, forecast_test = tts(
        forecast_in_copy, train_size=0.5, shuffle=False, stratify=None
    )

    target = "Target"
    d_train = lgb.Dataset(
        forecast_train.drop(columns=[target]), label=forecast_train[target]
    )

    params = {
        "boosting_type": "gbdt",
        "objective": "regression",
        "metric": "rmsle",
        "max_depth": 6,
        "learning_rate": 0.1,
        "verbose": 0,
        "num_threads": 16,
    }

    model = lgb.train(params, d_train, 100, verbose_eval=1)

    ensemble_lgb = pd.DataFrame(index=forecast_test.index)

    ensemble_lgb["ensemble_lgb"] = model.predict(forecast_test.drop(columns=[target]))

    ensemble_lgb_out = pd.DataFrame(index=forecast_out.index)

    ensemble_lgb_out["ensemble_lgb"] = model.predict(forecast_out)

    return ensemble_lgb, ensemble_lgb_out


def ensemble_tsfresh(forecast_in, forecast_out, season, perd):
    """
    Create rolled time series for ts feature extraction
    """

    def tsfresh_run(forecast, season, insample=True, forecast_out=None):
        df_roll_prep = forecast.reset_index()
        if insample:
            df_roll_prep = df_roll_prep.drop(["Target", "Date"], axis=1)
            df_roll_prep["id"] = 1
            target = forecast["Target"]
        else:
            df_roll_prep = df_roll_prep.drop(["index"], axis=1)
            df_roll_prep["id"] = 1

        df_roll = roll_time_series(
            df_roll_prep,
            column_id="id",
            column_sort=None,
            column_kind=None,
            rolling_direction=1,
            max_timeshift=season - 1,
        )
        counts = df_roll["id"].value_counts()
        df_roll_cut = df_roll[df_roll["id"].isin(counts[counts >= season].index)]

        # TS feature extraction
        concat_df = pd.DataFrame()

        concat_df = extract_features(
            df_roll_cut.ffill(),
            column_id="id",
            column_sort="sort",
            n_jobs=season,
            show_warnings=False,
            disable_progressbar=True,
        )

        if insample:

            concat_df = concat_df.dropna(axis=1, how="all")
            concat_df.index = (
                target[df_roll_cut["id"].value_counts().index]
                .sort_index()
                .to_frame()
                .index
            )
            concat_df = pd.merge(
                target[df_roll_cut["id"].value_counts().index].sort_index().to_frame(),
                concat_df,
                left_index=True,
                right_index=True,
                how="left",
            )
            concat_df_list = constant_feature_detect(data=concat_df, threshold=0.95)
            concat_df = concat_df.drop(concat_df_list, axis=1)
        else:
            forecast_out.index.name = "Date"
            concat_df.index = forecast_out.index

        concat_df = impute(concat_df)

        return concat_df

    _LOG.info("LightGBM ensemble have been successfully built")

    concat_df_drop_in = tsfresh_run(forecast_in, season, insample=True)

    extracted_n_selected = select_features(
        concat_df_drop_in.drop("Target", axis=1),
        concat_df_drop_in["Target"],
        fdr_level=0.01,
        n_jobs=12,
    )  # fdr is the significance level.

    forecast_out_add = pd.concat(
        (forecast_in.iloc[-season + 1 :, :].drop(["Target"], axis=1), forecast_out),
        axis=0,
    )
    concat_df_drop_out = tsfresh_run(
        forecast_out_add, season, insample=False, forecast_out=forecast_out
    )
    extracted_n_selected_out = concat_df_drop_out[extracted_n_selected.columns]

    # Reduce the dimensions of generated time series features
    pca2 = PCA(n_components=8)
    pca2.fit(extracted_n_selected)
    pca2_results_in = pca2.transform(extracted_n_selected)
    pca2_results_out = pca2.transform(extracted_n_selected_out)

    cols = 0
    for i in range(pca2_results_in.shape[1]):
        cols = cols + 1
        extracted_n_selected["pca_" + str(i)] = pca2_results_in[:, i]
        extracted_n_selected_out["pca_" + str(i)] = pca2_results_out[:, i]

    df = forecast_in.iloc[season - 1 :, :].copy()
    df = time_feature(df, perd)
    df["mean"] = df.drop(["Target"], axis=1).mean(axis=1)

    df_new = pd.concat(
        (df.reset_index(), extracted_n_selected.iloc[:, -cols:].reset_index(drop=True)),
        axis=1,
    )
    df_new = df_new.set_index("Date")
    forecast_train, forecast_test = tts(
        df_new, train_size=0.5, shuffle=False, stratify=None
    )
    target = "Target"
    d_train = lgb.Dataset(
        forecast_train.drop(columns=[target]), label=forecast_train[target]
    )

    params = {
        "boosting_type": "gbdt",
        "objective": "regression",
        "metric": "rmsle",
        "max_depth": 6,
        "learning_rate": 0.1,
        "verbose": 0,
        "num_threads": 16,
    }

    model = lgb.train(params, d_train, 100, verbose_eval=1)

    ensemble_ts = pd.DataFrame(index=forecast_test.index)

    ensemble_ts["ensemble_ts"] = model.predict(forecast_test.drop(columns=[target]))

    df_out = forecast_out.copy()
    df_out = time_feature(df_out, perd)
    df_out["mean"] = df_out.mean(axis=1)

    ensemble_ts_out = pd.DataFrame(index=df_out.index)
    ensemble_ts_out["ensemble_ts"] = model.predict(df_out)

    _LOG.info("LightGBM ensemble have been successfully built")

    return ensemble_ts, ensemble_ts_out


def ensemble_pure(forecast_in, forecast_out):
    """
    Pure Emsemble
    """

    df_perf = ensemble_performance(forecast_in).drop("Target", axis=1)

    def run_ensemble(df_perf, forecast):

        many = len(df_perf.iloc[0, :].sort_values())

        # Note these can fail, should see if that many indices actually exists.
        df_ensemble = pd.DataFrame(index=forecast.index)
        if many == 1:
            ValueError("You need more than one model to ensemble.")
        if many >= 2:
            df_ensemble[
                "_".join(list(df_perf.iloc[0, :].sort_values()[:2].index.values))
            ] = forecast[list(df_perf.iloc[0, :].sort_values()[:2].index)].mean(axis=1)
        if many >= 3:
            df_ensemble[
                "_".join(list(df_perf.iloc[0, :].sort_values()[:3].index.values))
            ] = forecast[list(df_perf.iloc[0, :].sort_values()[:3].index)].mean(axis=1)
        if many >= 4:
            df_ensemble[
                "_".join(list(df_perf.iloc[0, :].sort_values()[:4].index.values))
            ] = forecast[list(df_perf.iloc[0, :].sort_values()[:4].index)].mean(axis=1)

        return df_ensemble

    df_ensemble_in = run_ensemble(df_perf, forecast_in)
    df_ensemble_out = run_ensemble(df_perf, forecast_out)

    return df_ensemble_in, df_ensemble_out


def middle(ensemble_lgb, ensemble_ts, pure_ensemble):
    first_merge = pd.merge(
        ensemble_ts, ensemble_lgb, left_index=True, right_index=True, how="left"
    )
    second_merge = pd.merge(
        first_merge, pure_ensemble, left_index=True, right_index=True, how="left"
    )
    return second_merge


def ensemble_first(middle_in, forecast_in):
    third_merge = pd.merge(
        middle_in,
        forecast_in[["Target"]],
        left_index=True,
        right_index=True,
        how="left",
    )
    return third_merge, ensemble_performance(third_merge).drop("Target", axis=1)


def ensemble_doubled(middle_in, middle_out, forecast_in, forecast_out):

    third_merge_in = pd.merge(
        middle_in.drop(["Target"], axis=1),
        forecast_in,
        left_index=True,
        right_index=True,
        how="left",
    )
    third_merge_out = pd.merge(
        middle_out, forecast_out, left_index=True, right_index=True, how="left"
    )

    # Double Ensemble
    df_perf = ensemble_performance(third_merge_in).drop("Target", axis=1)

    def inner_ensemble(df_perf, third_merge):
        df_ensemble = pd.DataFrame(index=third_merge.index)
        # Note these can fail, should see if that many indices actually exists.

        many = len(df_perf.iloc[0, :].sort_values())

        if many == 1:
            ValueError("You need more than one model to ensemble.")
        if many >= 2:
            df_ensemble[
                "__X__".join(list(df_perf.iloc[0, :].sort_values()[:2].index.values))
            ] = third_merge[list(df_perf.iloc[0, :].sort_values()[:2].index)].mean(
                axis=1
            )
        if many >= 3:
            df_ensemble[
                "__X__".join(list(df_perf.iloc[0, :].sort_values()[:3].index.values))
            ] = third_merge[list(df_perf.iloc[0, :].sort_values()[:3].index)].mean(
                axis=1
            )
        if many >= 5:
            df_ensemble[
                "__X__".join(list(df_perf.iloc[0, :].sort_values()[:5].index.values))
            ] = third_merge[list(df_perf.iloc[0, :].sort_values()[:5].index)].mean(
                axis=1
            )
        if many >= 7:
            df_ensemble[
                "__X__".join(list(df_perf.iloc[0, :].sort_values()[:7].index.values))
            ] = third_merge[list(df_perf.iloc[0, :].sort_values()[:7].index)].mean(
                axis=1
            )
        return df_ensemble

    df_ensembled_in = inner_ensemble(df_perf, third_merge_in)
    df_ensembled_out = inner_ensemble(df_perf, third_merge_out)

    last_merge_in = pd.merge(
        third_merge_in, df_ensembled_in, left_index=True, right_index=True, how="left"
    )  # .drop(["month","mean"],axis=1)
    last_merge_out = pd.merge(
        third_merge_out, df_ensembled_out, left_index=True, right_index=True, how="left"
    )

    df_perf_last = ensemble_performance(last_merge_in).drop("Target", axis=1)

    return last_merge_in, last_merge_out, df_perf_last



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: Base Model"""


class BaseModel:
    """Base Model class of ForecastGA"""

    def __init__(self, config):

        if config.in_sample is None:
            raise ValueError(
                "The config class must be initialized with \
                              `set_in_sample()` or `set_out_sample()` prior to \
                              passing to a model."
            )

        self.seasonality = config.seasonality
        self.forecast_len = config.forecast_len
        self.freq = config.freq
        self.in_sample = config.in_sample
        self.GPU = config.GPU

        self.dataframe = config.dataframe
        self.train_df = config.train_df
        self.forecast_df = config.forecast_df
        self.seasons = config.seasons
        self.periods = config.periods
        self.model = None
        self.prediction = None

    def train(self):
        raise NotImplementedError

    def forecast(self):
        raise NotImplementedError



# encoding: utf-8

from forecastga import ga


class Blueprint:
    def __init__(self, description):
        self.raw = description
        self.scope = description.get("scope")
        self.defaults = description.get("defaults")
        self._identity = description.get("identity")
        self._queries = description.get("queries")

    @property
    def identity(self):
        data = self._identity
        if data:
            if isinstance(data, ga.utils.basestring):
                return dict(identity=data)
            if isinstance(data, dict):
                return data

        return None

    def queries(self, profile):
        base = ga.query.describe(profile, self.defaults)

        queries = []
        for title, description in self._queries.items():
            query = ga.query.refine(base, description)
            query.title = title
            queries.append(query)

        return queries



# encoding: utf-8

import functools
import inspector


class memoize:
    def __init__(self, function):
        self.function = function
        self.memoized = {}

    def __call__(self, *args):
        try:
            return self.memoized[args]
        except KeyError:
            self.memoized[args] = self.function(*args)
            return self.memoized[args]


def vectorize(fn):
    """
    Allows a method to accept one or more values,
    but internally deal only with a single item,
    and returning a list or a single item depending
    on what is desired.
    """

    @functools.wraps(fn)
    def vectorized_method(self, values, *vargs, **kwargs):
        wrap = not isinstance(values, (list, tuple))
        should_unwrap = not kwargs.setdefault("wrap", False)
        unwrap = wrap and should_unwrap
        del kwargs["wrap"]

        if wrap:
            values = [values]

        results = [fn(self, value, *vargs, **kwargs) for value in values]

        if unwrap:
            results = results[0]

        return results

    return vectorized_method


def immutable(method):
    @inspector.wraps(method)
    def wrapped_method(self, *vargs, **kwargs):
        obj = self.clone()
        method(obj, *vargs, **kwargs)
        return obj

    return wrapped_method


def identity(value):
    return value


def soak(*vargs, **kwargs):
    pass



#! /usr/bin/env python
# coding: utf-8
#

"""TBAT Model"""

from tbats import TBATS

from forecastga.models.base import BaseModel


class TBAT_Model(BaseModel):
    """TBAT Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        bat = TBATS(use_arma_errors=False, use_box_cox=True, use_trend=True)
        self.model = bat.fit(self.train_df)

    def forecast(self):
        self.prediction = self.model.forecast(self.forecast_len)



#! /usr/bin/env python
# coding: utf-8
#
"""ForecastGA: Colab Utils"""

import matplotlib.pyplot as plt


def plot_colab(df, title=None, dark_mode=False):

    from IPython.display import Markdown as md

    plt.style.use("default")

    def show_md(txt):
        display(md(txt))

    if dark_mode:
        # Good all around color library
        plt.style.use("seaborn-colorblind")
        plt.rcParams.update(
            {
                "lines.color": "#565555",
                "legend.edgecolor": "#818080",
                "legend.borderpad": 0.6,
                "text.color": "white",
                "axes.facecolor": "#383838",
                "axes.edgecolor": "#565555",
                "axes.grid": True,
                "axes.labelcolor": "white",
                "grid.color": "#565555",
                "xtick.color": "white",
                "ytick.color": "white",
                "figure.facecolor": "#383838",
                "savefig.facecolor": "white",
                "savefig.edgecolor": "white",
                "font.sans-serif": "Liberation Sans",
                "lines.linewidth": 2,
                "figure.figsize": [15, 10],
                "font.size": 16,
            }
        )

    else:
        plt.style.use("seaborn-colorblind")
        plt.rcParams.update(
            {
                "legend.borderpad": 0.6,
                "axes.grid": True,
                "font.sans-serif": "Liberation Sans",
                "lines.linewidth": 2,
                "figure.figsize": [15, 10],
                "font.size": 16,
            }
        )

    if title:
        show_md("## {}".format(title))

    df.plot()
    plt.show()



#! /usr/bin/env python
# coding: utf-8


"""ForecastGA: Main
"""


__version__ = "0.1.15"


from forecastga.helpers.ga_data import get_ga_data
from forecastga.auto import AutomatedModel
from forecastga.helpers.data import print_model_info
from forecastga.helpers.colab import plot_colab


def help():
    print("Welcome to ForecastGA")
    print()
    print("To use:")
    print()
    print("Find Model Info:")
    print("forecastga.print_model_info()")
    print()
    print("Initialize Model:")
    print()
    print(
        """
    Google Analytics:

        data = { 'client_id': '<google api client_id>',
                 'client_secret': '<google api client_secret>',
                 'ga_end_date': '2019-12-31',
                 'ga_metric': 'sessions',
                 'ga_segment': 'organic traffic',
                 'ga_start_date': '2018-01-01',
                 'ga_url': 'https://analytics.google.com/analytics/web/?authuser=2#/report-home/aXXXXXwXXXXXpXXXXXX',
                 'identity': '<google api identity>',
                 'omit_values_over': 2000000
                }

        model_list = ["TATS", "TBATS1", "TBATP1", "TBATS2", "ARIMA"]
        am = forecastga.AutomatedModel(data , model_list=model_list, forecast_len=30 )

         """
    )
    print()
    print(
        """
    Pandas DataFrame:

        # CSV with columns: Date and Sessions
        df = pd.read_csv('ga_sessions.csv')
        df.Date = pd.to_datetime(df.Date)
        df = df.set_index("Date")
        data = df.Sessions

        model_list = ["TATS", "TBATS1", "TBATP1", "TBATS2", "ARIMA"]
        am = forecastga.AutomatedModel(data , model_list=model_list, forecast_len=30 )

         """
    )
    print()
    print("Forecast Insample:")
    print("forecast_in, performance = am.forecast_insample()")
    print()
    print("Forecast Outsample:")
    print("forecast_out = am.forecast_outsample()")
    print()
    print("Ensemble Performance:")
    print(
        "all_ensemble_in, all_ensemble_out, all_performance = am.ensemble(forecast_in, forecast_out)"
    )
    print()
    print("Pretty Plot in Google Colab")
    print(
        'forecastga.plot_colab(forecast_in, title="Insample Forecast", dark_mode=True)'
    )



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: Data Helpers"""

import pandas as pd
import numpy as np

from seasonal.periodogram import periodogram
from scipy.signal import find_peaks
from sklearn.model_selection import train_test_split as tts

from forecastga.helpers.ssa import mySSA
from forecastga.helpers.logging import get_logger

_LOG = get_logger(__name__)

from forecastga.models import MODELS


def print_model_info():
    _ = [
        print(v["name"], ":", v["description"])
        for k, v in MODELS.items()
        if v["status"] == "active"
    ]


def constant_feature_detect(data, threshold=0.98):
    """detect features that show the same value for the
    majority/all of the observations (constant/quasi-constant features)

    Parameters
    ----------
    data : pd.Dataframe
    threshold : threshold to identify the variable as constant

    Returns
    -------
    list of variables names
    """

    data_copy = data.copy(deep=True)
    quasi_constant_feature = []
    for feature in data_copy.columns:
        predominant = (
            (data_copy[feature].value_counts() / np.float(len(data_copy)))
            .sort_values(ascending=False)
            .values[0]
        )
        if predominant >= threshold:
            quasi_constant_feature.append(feature)
    _LOG.info(len(quasi_constant_feature), " variables are found to be almost constant")
    return quasi_constant_feature


# More Diverse Selection For TBAT
def infer_seasonality_ssa(train, index=1):
    ssa = mySSA(train)
    ssa.embed(embedding_dimension=36, verbose=False)
    ssa.decompose(True)
    rec = ssa.view_reconstruction(
        ssa.Xs[index], names="Seasonality", return_df=True, plot=False
    )
    peaks, _ = find_peaks(
        rec.values.reshape(
            len(rec),
        ),
        height=0,
    )
    peak_diffs = [j - i for i, j in zip(peaks[:-1], peaks[1:])]
    seasonality = max(peak_diffs, key=peak_diffs.count)
    return seasonality


# Good First Selection
def infer_seasonality(train, index=0):  # skip the first one, normally
    interval, power = periodogram(train, min_period=4, max_period=None)
    try:
        season = int(
            pd.DataFrame([interval, power])
            .T.sort_values(1, ascending=False)
            .iloc[0, index]
        )
    except:
        _LOG.warning("Welch Season failed, defaulting to  SSA solution")
        season = int(infer_seasonality_ssa(train, index=1))
    return season


def infer_periodocity(train):
    perd = pd.infer_freq(train.index)
    if perd in ["MS", "M", "BM", "BMS"]:
        periodocity = 12
    elif perd in ["BH", "H"]:
        periodocity = 24
    elif perd == "B":
        periodocity = 5
    elif perd == "D":
        periodocity = 7
    elif perd in ["W", "W-SUN", "W-MON", "W-TUE", "W-WED", "W-THU", "W-FRI", "W-SAT"]:
        periodocity = 52
    elif perd in ["Q", "QS", "BQ", "BQS"]:
        periodocity = 4
    elif perd in ["A", "BA", "AS", "BAS"]:
        periodocity = 10
    elif perd in ["T", "min"]:
        periodocity = 60
    elif perd == "S":
        periodocity = 60
    elif perd in ["L", "ms"]:
        periodocity = 1000
    elif perd in ["U", "us"]:
        periodocity = 1000
    elif perd == "N":
        periodocity = 1000

    return periodocity


def select_seasonality(train, season):
    if season == "periodocity":
        seasonality = infer_periodocity(train)
    elif season == "infer_from_data":
        seasonality = infer_seasonality(train)
    return seasonality


def add_freq(idx, freq=None):
    """Add a frequency attribute to idx, through inference or directly.

    Returns a copy.  If `freq` is None, it is inferred.
    """
    idx = idx.copy()
    if freq is None:
        if idx.freq is None:
            freq = pd.infer_freq(idx)
        else:
            return idx
    idx.freq = pd.tseries.frequencies.to_offset(freq)
    if idx.freq is None:
        raise AttributeError(
            "no discernible frequency found to `idx`.  Specify"
            " a frequency string with `freq`."
        )
    return idx


def parse_data(df):
    if type(df) == pd.DataFrame:
        if df.shape[1] > 1:
            raise ValueError("The dataframe should only contain one target column")
    elif type(df) == pd.Series:
        df = df.to_frame()
    else:
        raise TypeError(
            "Please supply a pandas dataframe with one column or a pandas series"
        )
    try:
        df.index.date
    except AttributeError:
        raise TypeError("The index should be a datetype")

    if df.isnull().any().values[0]:
        raise ValueError(
            "The dataframe cannot have any null values, please interpolate"
        )
    try:
        df.columns = ["Target"]
    except:
        raise ValueError("There should only be one column")

    df.index = df.index.rename("Date")
    df.index = add_freq(df.index)

    _LOG.info(
        "The data has been successfully parsed by infering a frequency, and establishing a 'Date' index and 'Target' column."
    )

    return df, pd.infer_freq(df.index)


def train_test_split(df, forecast_len=30):

    train, test = tts(
        df["Target"], test_size=forecast_len, shuffle=False, stratify=None
    )
    _LOG.info(
        "An insample split of training size {} and testing size {} has been constructed".format(
            len(train), len(test)
        )
    )
    return train, test


def season_list(train):
    lista = []
    for i in range(15):
        i = 1 + i
        lista.append(infer_seasonality_ssa(train, i))
    return lista


def get_unique_N(iterable, N):
    """Yields (in order) the first N unique elements of iterable.
    Might yield less if data too short."""
    seen = set()
    for e in iterable:
        if e in seen:
            continue
        seen.add(e)
        yield e
        if len(seen) == N:
            _LOG.info(
                "The following set of plausible SSA seasonalities have been identified: {}".format(
                    seen
                )
            )
            return


# simple batcher.
def data_generator(x_full, y_full, bs):
    def split(arr, size):
        arrays = []
        while len(arr) > size:
            slice_ = arr[:size]
            arrays.append(slice_)
            arr = arr[size:]
        arrays.append(arr)
        return arrays

    while True:
        for rr in split((x_full, y_full), bs):
            yield rr



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: Prophet Model with Box-Cox transform of the data"""

import pandas as pd
from fbprophet import Prophet
from scipy.stats import boxcox
from scipy.special import inv_boxcox

from forecastga.models.base import BaseModel


class Prophet_BoxCox_Model(BaseModel):
    """Prophet Boxcox Model Class"""

    def __init__(self, config):
        self.boxcox_lambda = None
        super().__init__(config)

    def train(self, **kwargs):

        country_holidays = kwargs.get("country_holidays", "US")

        if self.freq == "D":
            ptm = Prophet(weekly_seasonality=True)
        else:
            ptm = Prophet()

        ptm.add_country_holidays(country_name=country_holidays)

        formatted_data = self.format_input(self.train_df)
        transformed_y, self.boxcox_lambda = boxcox(formatted_data["y"] + 1)
        formatted_data["y"] = transformed_y

        self.model = ptm.fit(formatted_data)

    def forecast(self):
        future = self.model.make_future_dataframe(
            periods=self.forecast_len, freq=self.freq
        )
        future_pred = self.model.predict(future)
        future_pred = future_pred[-self.forecast_len :]
        if self.boxcox_lambda:
            future_pred["yhat"] = (
                inv_boxcox(future_pred["yhat"], self.boxcox_lambda) - 1
            )
        self.prediction = self.format_output(future_pred)

    @staticmethod
    def format_input(df):
        df_pr = df.reset_index()
        df_pr.columns = ["ds", "y"]
        return df_pr

    @staticmethod
    def format_output(df):
        prophet_pred = pd.DataFrame({"Date": df["ds"], "Target": df["yhat"]})
        prophet_pred = prophet_pred.set_index("Date")
        return prophet_pred["Target"].values



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: HWAMS Model"""

from statsmodels.tsa.holtwinters import ExponentialSmoothing

from forecastga.models.base import BaseModel


class HWAMS_Model(BaseModel):
    """HWAMS Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):

        for i in range(3):

            params = [
                {"trend": "add", "seasonal": "mul", "use_boxcox": True},
                {"trend": "add", "seasonal": "mul", "use_boxcox": False},
                {"trend": None, "seasonal": "add", "use_boxcox": False},
            ]
            try:
                self.model = ExponentialSmoothing(
                    self.train_df,
                    seasonal_periods=self.seasons,
                    trend=params[i]["trend"],
                    seasonal=params[i]["seasonal"],
                    damped_trend=True,
                ).fit(use_boxcox=params[i]["use_boxcox"])
                break

            except:
                continue

    def forecast(self):
        self.prediction = self.model.forecast(self.forecast_len)



# encoding: utf-8

"""
Convenience functions for authenticating with Google
and asking for authorization with Google, with
`authenticate` at its core.

`authenticate` will do what it says on the tin, but unlike
the basic `googleanalytics.oauth.authenticate`, it also tries
to get existing credentials from the keyring, from environment
variables, it prompts for information when required and so on.
"""
import re

from . import oauth
from .oauth import Flow, Credentials


def navigate(
    accounts,
    account=None,
    webproperty=None,
    profile=None,
    ga_url=None,
    default_profile=True,
):

    if ga_url:
        return get_profile_from_url(accounts, ga_url)

    if webproperty and not account:
        raise KeyError(
            "Cannot navigate to a webproperty or profile without knowing the account."
        )
    if profile and not (webproperty and account):
        raise KeyError(
            "Cannot navigate to a profile without knowing account and webproperty."
        )

    if profile:
        return accounts[account].webproperties[webproperty].profiles[profile]
    elif webproperty:
        scope = accounts[account].webproperties[webproperty]
        if default_profile:
            return scope.profile
        else:
            return scope
    elif account:
        return accounts[account]
    else:
        return accounts


def get_profile_from_url(accounts, ga_url):

    if isinstance(ga_url, str) and "https://analytics.google.com/" in ga_url:

        psearch = re.search(
            "^https:\/\/analytics\.google\.com\/analytics\/web\/.*\/a(?P<a>[0-9]+)w(?P<w>[0-9]+)p(?P<p>[0-9]+).*$",
            str(ga_url),
            re.IGNORECASE,
        )

        if len(psearch.groups()) == 3:
            return get_profile(accounts, psearch["a"], psearch["w"], psearch["p"])

        else:
            error = "The URL was not correct.  it should include a portion matching `/a23337837w45733833p149423361/`"

    else:
        error = "The url provided should start with `https://analytics.google.com\/`"

    raise KeyError(error)


def get_profile(accounts, account, webproperty, profile):

    try:

        account = accounts[account]
        webproperty = [
            w
            for w in account.webproperties
            if w.raw["internalWebPropertyId"] == webproperty
        ][0]
        profile = webproperty.profiles[profile]

        return profile

    except Exception as e:
        print("Unknown Exception:", str(e))
        return None


def find(**kwargs):
    return oauth.Credentials.find(**kwargs)


def identity(name):
    return find(identity=name)


def authenticate(
    client_id=None,
    client_secret=None,
    client_email=None,
    private_key=None,
    access_token=None,
    refresh_token=None,
    account=None,
    webproperty=None,
    profile=None,
    ga_url=None,
    identity=None,
    prefix=None,
    suffix=None,
    interactive=False,
    save=False,
):
    """
    The `authenticate` function will authenticate the user with the Google Analytics API,
    using a variety of strategies: keyword arguments provided to this function, credentials
    stored in in environment variables, credentials stored in the keychain and, finally, by
    asking for missing information interactively in a command-line prompt.

    If necessary (but only if `interactive=True`) this function will also allow the user
    to authorize this Python module to access Google Analytics data on their behalf,
    using an OAuth2 token.
    """

    credentials = oauth.Credentials.find(
        valid=True,
        interactive=interactive,
        prefix=prefix,
        suffix=suffix,
        client_id=client_id,
        client_secret=client_secret,
        client_email=client_email,
        private_key=private_key,
        access_token=access_token,
        refresh_token=refresh_token,
        identity=identity,
    )

    if credentials.incomplete:
        if interactive:
            credentials = authorize(
                client_id=credentials.client_id,
                client_secret=credentials.client_secret,
                save=save,
                identity=credentials.identity,
                prefix=prefix,
                suffix=suffix,
            )
        elif credentials.type == 2:
            credentials = authorize(
                client_email=credentials.client_email,
                private_key=credentials.private_key,
                identity=credentials.identity,
                save=save,
            )
        else:
            raise KeyError(
                "Cannot authenticate: enable interactive authorization, pass a token or use a service account."
            )

    accounts = oauth.authenticate(credentials)
    scope = navigate(
        accounts,
        account=account,
        webproperty=webproperty,
        profile=profile,
        ga_url=ga_url,
    )
    return scope


def authorize(
    client_id=None,
    client_secret=None,
    client_email=None,
    private_key=None,
    save=False,
    identity=None,
    prefix=None,
    suffix=None,
):
    base_credentials = oauth.Credentials.find(
        valid=True,
        interactive=True,
        identity=identity,
        client_id=client_id,
        client_secret=client_secret,
        client_email=client_email,
        private_key=private_key,
        prefix=prefix,
        suffix=suffix,
    )

    if base_credentials.incomplete:
        credentials = oauth.authorize(
            base_credentials.client_id, base_credentials.client_secret
        )
        credentials.identity = base_credentials.identity
    else:
        credentials = base_credentials

    return credentials


def revoke(
    client_id,
    client_secret,
    client_email=None,
    private_key=None,
    access_token=None,
    refresh_token=None,
    identity=None,
    prefix=None,
    suffix=None,
):

    """
    Given a client id, client secret and either an access token or a refresh token,
    revoke OAuth access to the Google Analytics data and remove any stored credentials
    that use these tokens.
    """

    if client_email and private_key:
        raise ValueError("Two-legged OAuth does not use revokable tokens.")

    credentials = oauth.Credentials.find(
        complete=True,
        interactive=False,
        identity=identity,
        client_id=client_id,
        client_secret=client_secret,
        access_token=access_token,
        refresh_token=refresh_token,
        prefix=prefix,
        suffix=suffix,
    )

    retval = credentials.revoke()
    return retval



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['forecastga',
 'forecastga.ga',
 'forecastga.ga.auth',
 'forecastga.ga.utils',
 'forecastga.helpers',
 'forecastga.models']

package_data = \
{'': ['*'], 'forecastga': ['stan/unix/*', 'stan/win/*']}

install_requires = \
['MarkupSafe>=1.1.1,<2.0.0',
 'PyYAML>=5.3.1,<6.0.0',
 'Werkzeug>=1.0.1,<2.0.0',
 'addressable>=1.4.2,<2.0.0',
 'colorama>=0.4.4,<0.5.0',
 'fbprophet>=0.7.1,<0.8.0',
 'gluonts>=0.6.4,<0.7.0',
 'google-api-core>=1.24.1,<2.0.0',
 'google-api-python-client>=1.12.8,<2.0.0',
 'google-auth-httplib2>=0.0.4,<0.0.5',
 'google-auth>=1.24.0,<2.0.0',
 'holidays>=0.10.4,<0.11.0',
 'httplib2>=0.18.1,<0.19.0',
 'inspect-it>=0.3.2,<0.4.0',
 'keyring>=21.8.0,<22.0.0',
 'lightgbm>=3.1.1,<4.0.0',
 'matplotlib==3.2.2',
 'mxnet>=1.7.0,<2.0.0',
 'nbeats-pytorch>=1.3.1,<2.0.0',
 'numpy==1.19.3',
 'oauth2client>=4.1.3,<5.0.0',
 'pandas==1.0.5',
 'pmdarima>=1.8.0,<2.0.0',
 'prettytable>=2.0.0,<3.0.0',
 'pylev>=1.3.0,<2.0.0',
 'pystan>=2.19.1,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.25.1,<3.0.0',
 'rsa>=4.6,<5.0',
 'scikit-learn>=0.24.0,<0.25.0',
 'scipy==1.5.0',
 'seasonal>=0.3.1,<0.4.0',
 'setuptools>=51.1.1,<52.0.0',
 'snakify>=1.1.1,<2.0.0',
 'statsmodels>=0.12.1,<0.13.0',
 'tbats>=1.1.0,<2.0.0',
 'torch>=1.7.1,<2.0.0',
 'tqdm>=4.55.1,<5.0.0',
 'tsfresh>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'forecastga',
    'version': '0.1.15',
    'description': 'A Python tool to forecast GA data using several popular timeseries models',
    'long_description': '# ForecastGA\nA Python tool to forecast GA data using several popular timeseries models.\n\n[![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/drive/1nmcu37MY02dfMdUbinrwwg7gA9ya3eud?usp=sharing)\n\n\n## About\n\n### Welcome to ForecastGA\n\nForecastGA is a tool that combines a couple of popular libraries, [Atspy](https://github.com/firmai/atspy) and [googleanalytics](https://github.com/debrouwere/google-analytics), with a few enhancements.\n\n* The models are made more intuitive to upgrade and add by having the tool logic separate from the model training and prediction.\n* When calling `am.forecast_insample()`, any kwargs included (e.g. `learning_rate`) are passed to the train method of the model.\n* Google Analytics profiles are specified by simply passing the URL (e.g. https://analytics.google.com/analytics/web/?authuser=2#/report-home/aXXXXXwXXXXXpXXXXXX).\n* You can provide a `data` dict with GA config options or a Pandas Series as the input data.\n* Multiple log levels.\n* Auto GPU detection (via Torch).\n* List all available models, with descriptions, by calling `forecastga.print_model_info()`.\n* Google API info can be passed in the `data` dict or uploaded as a JSON file named `identity.json`.\n* Created a companion Google Colab notebook to easily run on GPU.\n* A handy plot function for Colab, `forecastga.plot_colab(forecast_in, title="Insample Forecast", dark_mode=True)` that formats nicely and also handles Dark Mode!\n\n### Models Available\n* `ARIMA` : Automated ARIMA Modelling\n* `Prophet` : Modeling Multiple Seasonality With Linear or Non-linear Growth\n* `ProphetBC` : Prophet Model with Box-Cox transform of the data\n* `HWAAS` : Exponential Smoothing With Additive Trend and Additive Seasonality\n* `HWAMS` : Exponential Smoothing with Additive Trend and Multiplicative Seasonality\n* `NBEATS` : Neural basis expansion analysis (now fixed at 20 Epochs)\n* `Gluonts` : RNN-based Model (now fixed at 20 Epochs)\n* `TATS` : Seasonal and Trend no Box Cox\n* `TBAT` : Trend and Box Cox\n* `TBATS1` : Trend, Seasonal (one), and Box Cox\n* `TBATP1` : TBATS1 but Seasonal Inference is Hardcoded by Periodicity\n* `TBATS2` : TBATS1 With Two Seasonal Periods\n\n\n### How To Use\n\n#### Find Model Info:\n`forecastga.print_model_info()`\n\n#### Initialize Model:\n\n##### Google Analytics:\n\n```\ndata = { \'client_id\': \'<google api client_id>\',\n         \'client_secret\': \'<google api client_secret>\',\n         \'identity\': \'<google api identity>\',\n         \'ga_start_date\': \'2018-01-01\',\n         \'ga_end_date\': \'2019-12-31\',\n         \'ga_metric\': \'sessions\',\n         \'ga_segment\': \'organic traffic\',\n         \'ga_url\': \'https://analytics.google.com/analytics/web/?authuser=2#/report-home/aXXXXXwXXXXXpXXXXXX\',\n         \'omit_values_over\': 2000000\n        }\n\nmodel_list = ["TATS", "TBATS1", "TBATP1", "TBATS2", "ARIMA"]\nam = forecastga.AutomatedModel(data , model_list=model_list, forecast_len=30 )\n```\n\n##### Pandas DataFrame:\n\n```\n# CSV with columns: Date and Sessions\ndf = pd.read_csv(\'ga_sessions.csv\')\ndf.Date = pd.to_datetime(df.Date)\ndf = df.set_index("Date")\ndata = df.Sessions\n\nmodel_list = ["TATS", "TBATS1", "TBATP1", "TBATS2", "ARIMA"]\nam = forecastga.AutomatedModel(data , model_list=model_list, forecast_len=30 )\n```\n\n#### Forecast Insample:\n`forecast_in, performance = am.forecast_insample()`\n\n#### Forecast Outsample:\n`forecast_out = am.forecast_outsample()`\n\n#### Ensemble Performance:\n`all_ensemble_in, all_ensemble_out, all_performance = am.ensemble(forecast_in, forecast_out)`\n\n#### Pretty Plot in Google Colab\n`forecastga.plot_colab(forecast_in, title="Insample Forecast", dark_mode=True)`\n\n\n# Installation\nWindows users may need to manually install the two items below via conda :\n1. `conda install pystan`\n1. `conda install pytorch -c pytorch`\n1. `!pip install --upgrade git+https://github.com/jroakes/ForecastGA.git`\n\notherwise,\n`pip install --upgrade forecastga`\n\nThis repo support GPU training. Below are a few libraries that may have to be manually installed to support.\n```\npip install --upgrade mxnet-cu101\npip install --upgrade torch 1.7.0+cu101\n```\n\n\n## Acknowledgements\n\n1. Majority of forecasting code taken from https://github.com/firmai/atspy and refactored heavily.\n1. Google Analytics based off of: https://github.com/debrouwere/google-analytics\n1. Thanks to [richardfergie](https://github.com/richardfergie) for the addition of the Prophet Box-Cox model to control negative predictions.\n\n## Contribute\nThe goal of this repo is to grow the list of available models to test.  If you would like to contribute one please read on.  Feel free to have fun naming your models.\n\n1. Fork the repo.\n2. In the `/src/forecastga/models` folder there is a model called `template.py`.  You can use this as a template for creating your new model.  All available variables are there. Forecastga ensures each model has the right data and calls only the `train` and `forecast` methods for each model. Feel free to add additional methods that your model requires.\n3. Edit the `/src/forecastga/models/__init__.py` file to add your model\'s information.  Follow the format of the other entries.  Forecastga relies on `loc` to find the model and `class` to find the class to use.\n4. Edit `requirments.txt` with any additional libraries needed to run your model.  Keep in mind that this repo should support GPU training if available and some libraries have separate GPU-enabled versions.\n5. Issue a pull request.\n\nIf you enjoyed this tool consider buying me some beer at: [Paypalme](https://www.paypal.com/paypalme/codeseo)\n',
    'author': 'JR Oakes',
    'author_email': 'jroakes@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jroakes/ForecastGA',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



#! /usr/bin/env python
# coding: utf-8

"""ForecastGA: Google Analytics Helper Functions"""

import os
from types import SimpleNamespace
from datetime import datetime
import pandas as pd
import numpy as np

from forecastga.helpers.logging import get_logger

from forecastga import ga

_LOG = get_logger(__name__)


def load_identity(data=None):

    if data:
        jf = {
            k.lower(): v
            for k, v in data.items()
            if k.lower() in ["client_id", "client_secret", "identity"]
        }
        return SimpleNamespace(**jf)

    if not os.path.isfile("identity.json"):
        raise FileExistsError(
            "A JSON file named `identity.json` must be accessible with your API credentials."
        )

    with open("identity.json") as f:
        jf = json.load(f)
        identify_json = SimpleNamespace(**jf)

    return identify_json


def load_profile(ga_url, identify_ns):

    try:
        profile = ga.authenticate(
            client_id=identify_ns.client_id,
            client_secret=identify_ns.client_secret,
            identity=identify_ns.identity,
            ga_url=ga_url,
            interactive=True,
        )
        _LOG.info("Authenticated")
        return profile

    except Exception as e:
        _LOG.error("An error occured: " + str(e))
        return None


def p_date(_dt):
    return datetime.strftime(_dt, "%Y-%m-%d")


def get_ga_data(data):

    if "client_id" in data and "client_secret" in data and "identity" in data:
        identify_ns = load_identity(data)
    else:
        identify_ns = load_identity()

    if "ga_url" not in data:
        raise AttributeError(
            "You must provide the URL for your Google Analytics property."
        )

    profile = load_profile(data["ga_url"], identify_ns)

    if profile is None:
        return None

    data = SimpleNamespace(**data)

    try:
        print(
            "Pulling data from {} to {}.".format(data.ga_start_date, data.ga_end_date)
        )
        sessions = (
            profile.core.query.metrics(data.ga_metric)
            .segment(data.ga_segment)
            .daily(data.ga_start_date, data.ga_end_date)
            .report
        )

    except Exception as e:
        _LOG.error("Error. Error retreiving data from Google Analytics.", str(e))
        return None

    df = sessions.as_dataframe()

    df["date"] = pd.to_datetime(df["date"])

    # Clean data.
    if data.omit_values_over and int(data.omit_values_over) > 0:
        df.loc[df[data.ga_metric] > data.omit_values_over, data.ga_metric] = np.nan

    df.loc[df[data.ga_metric] < 1, data.ga_metric] = np.nan

    df.dropna(inplace=True, axis=0)

    _LOG.info(
        "Rows: {rows} Min Date: {min_date} Max Date: {max_date}".format(
            rows=len(df), min_date=p_date(df.date.min()), max_date=p_date(df.date.max())
        )
    )
    # Backfilling missing values
    df = df.set_index("date").asfreq("d", method="bfill")

    return df[data.ga_metric]



#! /usr/bin/env python
# coding: utf-8


"""ForecastGA: AutoModel"""

import importlib
import warnings
import torch
import pandas as pd
from statsmodels.tools.eval_measures import rmse

from forecastga.helpers.logging import get_logger
from forecastga.helpers.data import parse_data, train_test_split, select_seasonality
from forecastga.ensembles import (
    ensemble_lightgbm,
    ensemble_tsfresh,
    ensemble_pure,
    middle,
    ensemble_first,
    ensemble_doubled,
)
from forecastga.models import MODELS
from forecastga.helpers.ga_data import get_ga_data

pd.plotting.register_matplotlib_converters()
warnings.filterwarnings("ignore")

_LOG = get_logger(__name__)


class ModelConfig:
    def __init__(
        self,
        df: pd.Series,
        seasonality: str = "infer_from_data",
        forecast_len: int = 20,
        GPU: bool = torch.cuda.is_available(),
    ):

        self.df = df
        self.seasonality = seasonality
        self.forecast_len = forecast_len
        self.GPU = GPU

        self.in_sample = None
        self.train_df = None
        self.forecast_df = None
        self.seasons = None
        self.periods = None

        self.dataframe, self.freq = parse_data(df)

    def set_in_sample(self):

        self.in_sample = True

        self.train_df, self.forecast_df = train_test_split(
            self.dataframe, forecast_len=self.forecast_len
        )
        self.forecast_len = len(self.forecast_df)

        self.seasons = select_seasonality(self.train_df, self.seasonality)
        self.periods = select_seasonality(self.train_df, "periodocity")

    def set_out_sample(self):

        self.in_sample = False

        self.train_df, self.forecast_df = self.dataframe["Target"], None

        self.seasons = select_seasonality(self.train_df, self.seasonality)
        self.periods = select_seasonality(self.train_df, "periodocity")


class AutomatedModel:
    def __init__(
        self,
        data,
        model_list: list = [],
        seasonality: str = "infer_from_data",
        forecast_len: int = 20,
        GPU: bool = torch.cuda.is_available(),
    ):

        if isinstance(data, dict):
            self.df = get_ga_data(data)
        elif isinstance(data, pd.Series):
            self.df = data
        else:
            raise AttributeError(
                "First parameter must be a dict with GA property and time data, or a pandas series."
            )
        self.model_list = model_list
        self.seasonality = seasonality
        self.forecast_len = forecast_len
        self.GPU = GPU
        self.models_dict = {}
        self.forecast_dict = {}

        self.config: ModelConfig = ModelConfig(
            self.df,
            seasonality=seasonality,
            forecast_len=forecast_len,
            GPU=GPU,
        )

    def forecast_insample(self, **kwargs):

        self.config.set_in_sample()

        self.models_dict = self.__train_models(**kwargs)
        self.forecast_dict = self.__forecast_models()
        forecast_frame = self.forecast_dataframe(
            self.config.forecast_df, self.forecast_dict
        )
        preformance = self.insample_performance(forecast_frame)

        _LOG.info("Successfully finished in sample forecast")

        return forecast_frame, preformance

    def forecast_outsample(self, **kwargs):

        self.config.set_out_sample()

        self.models_dict = self.__train_models(**kwargs)
        self.forecast_dict = self.__forecast_models()

        future_index = pd.date_range(
            self.config.dataframe.index[-1],
            periods=self.config.forecast_len + 1,
            freq=self.config.freq,
        )[1:]

        forecast_frame = self.forecast_dataframe(
            pd.DataFrame({"Target": 0}, index=future_index), self.forecast_dict
        )

        _LOG.info("Successfully finished out of sample forecast")

        return forecast_frame

    def available_models(self):
        return [k for k, v in MODELS.items() if v["status"] == "active"]

    def __train_models(self, **kwargs):

        models_dict = {}

        for model_name in self.model_list:
            if model_name not in self.available_models():
                _LOG.warning(
                    "Model {} is not available.  Skipping...".format(model_name)
                )
                continue

            _LOG.info(
                "Model {} is being loaded and trained for {} prediction".format(
                    model_name,
                    "in sample" if self.config.in_sample else "out of sample",
                )
            )
            model_data = MODELS[model_name]
            module = importlib.import_module(model_data["loc"])
            module_class = getattr(module, model_data["class"])
            model = module_class(self.config)

            model.train(**kwargs)

            models_dict[model_name] = model

        return models_dict

    def __forecast_models(self, models_dict=None):

        models_dict = models_dict or self.models_dict

        forecast_dict = {}
        for model_name, model in models_dict.items():

            _LOG.info(
                "Model {} is being used to forecast {}".format(
                    model_name,
                    "in sample" if self.config.in_sample else "out of sample",
                )
            )

            model.forecast()

            forecast_dict[model_name] = model.prediction

        return forecast_dict

    @staticmethod
    def forecast_dataframe(df, forecast_dict):
        insample = df.to_frame()
        for name, forecast in forecast_dict.items():
            insample[name] = forecast
        return insample

    @staticmethod
    def insample_performance(forecast_frame, as_dict=False):

        dict_perf = {}
        for col, _ in forecast_frame.iteritems():
            dict_perf[col] = {}
            dict_perf[col]["rmse"] = rmse(forecast_frame["Target"], forecast_frame[col])
            dict_perf[col]["mse"] = dict_perf[col]["rmse"] ** 2
            dict_perf[col]["mean"] = forecast_frame[col].mean()
        if as_dict:
            return dict_perf

        return pd.DataFrame.from_dict(dict_perf)

    def ensemble(self, forecast_in, forecast_out):
        season = self.seasonality

        _LOG.info("Building LightGBM Ensemble from TS data (ensemble_lgb)")

        ensemble_lgb_in, ensemble_lgb_out = ensemble_lightgbm(
            forecast_in, forecast_out, self.freq
        )

        _LOG.info(
            "Building LightGBM Ensemble from PCA reduced TSFresh Features (ensemble_ts). This can take a long time."
        )

        ensemble_ts_in, ensemble_ts_out = ensemble_tsfresh(
            forecast_in, forecast_out, season, self.freq
        )

        _LOG.info("Building Standard First Level Ensemble")
        df_ensemble_in, df_ensemble_out = ensemble_pure(forecast_in, forecast_out)
        middle_out = middle(ensemble_lgb_out, ensemble_ts_out, df_ensemble_out)
        middle_in = middle(ensemble_lgb_in, ensemble_ts_in, df_ensemble_in)

        _LOG.info("Building Final Multi-level Ensemble")
        middle_in, _ = ensemble_first(middle_in, forecast_in)
        all_ensemble_in, all_ensemble_out, all_performance = ensemble_doubled(
            middle_in, middle_out, forecast_in, forecast_out
        )

        return all_ensemble_in, all_ensemble_out, all_performance.T.sort_values("rmse")



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: TBATP1 Model (TBATS1 but Seasonal Inference is Hardcoded by Periodicity)"""

from tbats import TBATS

from forecastga.models.base import BaseModel


class TBATP1_Model(BaseModel):
    """TBATP1 Model Class"""

    def __init__(self, config):
        super().__init__(config)

    def train(self, **kwargs):
        bat = TBATS(
            seasonal_periods=[self.periods],
            use_arma_errors=False,
            use_box_cox=True,
            use_trend=True,
        )
        self.model = bat.fit(self.train_df)

    def forecast(self):
        self.prediction = self.model.forecast(self.forecast_len)



# encoding: utf-8

"""
"""

import collections
import csv
from datetime import datetime
import hashlib
import json
import time
from copy import deepcopy
from functools import partial

import addressable
import inspector
import yaml
from dateutil.relativedelta import relativedelta
import prettytable

from . import errors, utils
from .columns import Column, ColumnList, Segment


INTERVAL_TIMEDELTAS = {
    "year": dict(years=1),
    "year_month": dict(months=1),
    "year_week": dict(weeks=1),
    "date": dict(days=1),
    "date_hour": dict(hours=1),
}


def path(l, *keys):
    indexed = {}
    for el in l:
        branch = indexed
        for key in keys[:-1]:
            value = getattr(el, key)
            branch = branch.setdefault(value, {})
        value = getattr(el, keys[-1])
        branch.setdefault(value, el)
    return indexed


def default(metric):
    if "avg" in metric:
        return None
    else:
        return 0


class Report(object):
    """
    Executing a query will return a report, which contains the requested data.

    Queries are executed and turned into a report lazily, whenever data is requested.
    You can also explicitly generate a report from a query by using the `Query#get` method.

    ```python
    # will return a query object
    profile.core.query.metrics('pageviews').range('yesterday')
    # will return a report object
    profile.core.query.metrics('pageviews').range('yesterday').get()
    # will generate a report object and return its rows -- these
    # two are equivalent
    profile.core.query.metrics('pageviews').range('yesterday').rows
    profile.core.query.metrics('pageviews').range('yesterday').get().rows
    ```

    You can access the data in a Report object both rowwise and columnwise.

    ```python
    report = query.metrics('pageviews', 'sessions').range('yesterday')
    # first ten rows
    report.rows[:10]
    # work with just session data points
    report['sessions'][:10]
    report.rows[:10]['sessions']
    ```

    For simple data structures, there are also some shortcuts.

    These shortcuts are available both directly on Report objects
    and lazily-loaded via Query objects.

    ```python
    # reports with a single value
    query = profile.core.query('pageviews').range('yesterday')
    report = query.get()
    assert query.value == report.value
    # reports with a single metric
    profile.core.query('pageviews').daily('yesterday', days=-10).values
    # reports with a single result
    query = profile.core.query(['pageviews', 'sessions']).range('yesterday')
    assert query.first == query.last
    ```
    """

    def __init__(self, raw, query):
        self.raw = []
        self.queries = []

        all_columns = query.api.all_columns
        report_columns = [column["name"] for column in raw["columnHeaders"]]
        self.columns = ColumnList([all_columns[column] for column in report_columns])
        self.metrics = addressable.filter(
            lambda column: column.type == "metric", self.columns
        )
        self.dimensions = addressable.filter(
            lambda column: column.type == "dimension", self.columns
        )
        time_columns = ["date_hour", "date", "year_week", "year_month", "year"]
        try:
            self.granularity = next(
                column
                for column in self.dimensions
                if column.python_slug in time_columns
            )
        except StopIteration:
            self.granularity = None
        slugs = [column.python_slug for column in self.columns]
        self.Row = collections.namedtuple("Row", slugs)
        self.rows = []
        self.append(raw, query)

        self.since = self.until = None

        if "start-date" in raw["query"]:
            self.since = datetime.strptime(raw["query"]["start-date"], "%Y-%m-%d")

        if "end-date" in raw["query"]:
            self.until = datetime.strptime(raw["query"]["end-date"], "%Y-%m-%d")

    def append(self, raw, query):
        self.raw.append(raw)
        self.queries.append(query)
        self.is_complete = not "nextLink" in raw

        casters = [column.cast for column in self.columns]

        # if no rows were returned, the GA API doesn't
        # include the `rows` key at all
        for row in self.raw[-1].get("rows", []):
            typed_row = [casters[i](row[i]) for i in range(len(self.columns))]
            typed_tuple = self.Row(*typed_row)
            self.rows.append(typed_tuple)

        # TODO: figure out how this works with paginated queries
        self.totals = raw["totalsForAllResults"]
        # more intuitive when querying for just a single metric
        self.total = list(raw["totalsForAllResults"].values())[0]

    @property
    def first(self):
        if len(self.rows) == 0:
            return None
        else:
            return self.rows[0]

    @property
    def last(self):
        if len(self.rows) == 0:
            return None
        else:
            return self.rows[-1]

    @property
    def value(self):
        if len(self.rows) == 0:
            return None
        elif len(self.rows) == 1:
            return self.values[0]
        else:
            raise ValueError(
                "This report contains multiple rows or metrics. Please use `rows`, `first`, `last` or a column name."
            )

    @property
    def values(self):
        if len(self.metrics) == 1:
            metric = self.metrics[0]
            return self[metric]
        else:
            raise ValueError(
                "This report contains multiple metrics. Please use `rows`, `first`, `last` or a column name."
            )

    # TODO: it'd be nice to have metadata from `ga` available as
    # properties, rather than only having them in serialized form
    # (so e.g. the actual metric objects with both serialized name, slug etc.)
    def __expand(self):
        import ranges

        # 1. generate grid

        # 1a. generate date grid
        since = self.since
        until = self.until + relativedelta(days=1)
        granularity = self.granularity.slug
        dimensions = [column.slug for column in self.dimensions]
        metrics = [column.slug for column in self.metrics]

        time_step = INTERVAL_TIMEDELTAS[self.granularity.slug]
        time_ix = report.columns.index(self.granularity)
        time_interval = (since, until)
        time_range = list(ranges.date.range(*time_interval, **time_step))

        # 1b. generate dimension factor grid
        factors = {
            dimension: set(report[dimension])
            for dimension in set(dimensions) - {granularity}
        }

        # 1c. assemble grid
        grid = list(itertools.product(time_range, *factors.values()))

        # 2. fill in the grid with available data
        ndim = len(self.dimensions)
        index = path(report.rows, *report.slugs[:ndim])

        filled = []

        for row in grid:
            try:
                # navigate to data (if it exists)
                index_columns = row[:ndim]
                data = index
                for column in index_columns:
                    data = data[column]
                filled.append(data)
            except KeyError:
                # TODO: for some metrics (in particular averages)
                # the default value should be None, for most others
                # it should be zero
                row_metrics = [None] * len(report.metrics)
                filler = list(row) + row_metrics
                row = report.Row(*filler)
                filled.append(row)

        report.rows = filled
        return report

    def serialize(self, format=None, with_metadata=False):
        names = [column.name for column in self.columns]

        if not format:
            return self.as_dict(with_metadata=with_metadata)
        elif format == "json":
            return json.dumps(self.as_dict(with_metadata=with_metadata), indent=4)
        elif format == "csv":
            buf = utils.StringIO()
            writer = csv.writer(buf)
            writer.writerow(names)
            writer.writerows(self.rows)
            return buf.getvalue()
        elif format == "ascii":
            table = prettytable.PrettyTable(names)
            table.align = "l"
            for row in self.rows:
                table.add_row(row)
            if with_metadata:
                return utils.format(
                    """
                    {title}
                    {table}
                    """,
                    title=self.queries[0].title,
                    table=table,
                )
            else:
                return table

    def as_dict(self, with_metadata=False):
        serialized = []
        for row in self.rows:
            row = row._asdict()
            for key, value in row.items():
                row[key] = utils.date.serialize(value)
            serialized.append(row)

        if with_metadata:
            return {
                "title": self.queries[0].title,
                "queries": self.queries,
                "metrics": [column.name for column in self.metrics],
                "dimensions": [column.name for column in self.dimensions],
                "results": serialized,
            }
        else:
            return serialized

    def as_dataframe(self):
        import pandas

        # passing every row as a dictionary is not terribly efficient,
        # but it works for now
        return pandas.DataFrame(self.as_dict())

    def __getitem__(self, key):
        try:
            if isinstance(key, Column):
                key = key.slug
            i = self.columns.index(key)
            return [row[i] for row in self.rows]
        except ValueError:
            raise ValueError(key + " not in column headers")

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        return len(self.rows)

    # TODO: would be cool if we could split up headers
    # into metrics vs. dimensions so we could say
    # "pageviews by day, browser"
    # (also see `title` and `description` on query objects)
    def __repr__(self):
        metrics = ", ".join([header.name for header in self.metrics])
        dimensions = ",".join([header.name for header in self.dimensions])
        if len(dimensions):
            return "<googleanalytics.query.Report object: {} by {}".format(
                metrics, dimensions
            )
        else:
            return "<googleanalytics.query.Report object: {}".format(metrics)


EXCLUSION = {
    "eq": "neq",
    "neq": "eq",
    "gt": "lte",
    "lt": "gte",
    "gte": "lt",
    "lte": "gt",
    "re": "nre",
    "nre": "re",
    "contains": "ncontains",
    "ncontains": "contains",
}


def select(source, selection, invert=False):
    selections = []
    for key, values in selection.items():
        if "__" in key:
            column, method = key.split("__")
        else:
            column = key
            method = "eq"

        if not hasattr(Column, method):
            raise ValueError(
                "{method} is not a valid selector. Choose from: {options}".format(
                    method=method,
                    options=", ".join(Column.selectors),
                )
            )

        if invert:
            method = EXCLUSION[method]

        column = source[column]
        selector = getattr(column, method)

        if not isinstance(values, (list, tuple)):
            values = [values]

        # e.g. source=['cpc', 'cpm'] will return an OR
        # filter for these two sources
        for value in values:
            selections.append(selector(value))

    return selections


# TODO: refactor `raw` into `parameters`, with serialization
# (stringification) happening in one place, in `Query#build`
# and removing empty keys as necessary
# TODO: consider whether to pass everything through `Query#set`
# or otherwise avoid having two paths to modifying `raw`
class Query(object):
    """
    Return a query for certain metrics and dimensions.

    ```python
    # pageviews (metric) as a function of geographical region
    profile.core.query('pageviews', 'region')
    # pageviews as a function of browser
    profile.core.query(['pageviews'], ['browser'])
    ```

    The returned query can then be further refined using
    all methods available on the `CoreQuery` object, such as
    `limit`, `sort`, `segment` and so on.

    Metrics and dimensions may be either strings (the column id or
    the human-readable column name) or Metric or Dimension
    objects.

    Metrics and dimensions specified as a string are not case-sensitive.

    ```python
    profile.query('PAGEVIEWS')
    ```

    If specifying only a single metric or dimension, you can
    but are not required to wrap it in a list.
    """

    _lock = 0

    def __init__(self, api, parameters={}, metadata={}, title=None):
        self._title = title
        self.raw = {
            "ids": "ga:" + api.profile.id,
            "metrics": [],
            "dimensions": [],
        }
        self.raw.update(parameters)
        self.meta = {}
        self.meta.update(metadata)
        self.api = api
        self.profile = api.profile
        self.webproperty = api.profile.webproperty
        self.account = api.profile.webproperty.account
        self._report = None

    # do not execute more than one query per second
    def _wait(self):
        now = time.time()
        elapsed = now - self._lock
        wait = max(0, 1 - elapsed)
        time.sleep(wait)
        self._lock = time.time()
        return wait

    @property
    def endpoint(self):
        return self.account.service.data().ga()

    def clone(self):
        query = self.__class__(
            api=self.api,
            parameters=deepcopy(self.raw),
            metadata=deepcopy(self.meta),
        )
        return query

    @utils.immutable
    def set(self, key=None, value=None, **kwargs):
        """
        `set` is a way to add raw properties to the request,
        for features that this module does not
        support or supports incompletely. For convenience's
        sake, it will serialize Column objects but will
        leave any other kind of value alone.
        """

        serialize = partial(self.api.columns.serialize, greedy=False)

        if key and value:
            self.raw[key] = serialize(value)
        elif key or kwargs:
            properties = key or kwargs
            for key, value in properties.items():
                self.raw[key] = serialize(value)
        else:
            raise ValueError(
                "Query#set requires a key and value, a properties dictionary or keyword arguments."
            )

        return self

    @utils.immutable
    def columns(self, required_type=None, *values):
        for column in self.api.columns.normalize(values, wrap=True):
            if required_type and required_type != column.type:
                raise ValueError(
                    "Tried to add {type} but received: {column}".format(
                        type=required_type,
                        column=column,
                    )
                )
            self.raw[column.type + "s"].append(column.id)
        return self

    # TODO: maybe do something smarter, like {granularity} {metrics}
    # by {dimensions} for {segment}, filtered by {filters}.
    # First {limit} results from {start} to {end} /
    # for {start=end}, sorted by {direction} {sort}.
    @property
    def description(self):
        """
        A list of the metrics this query will ask for.
        """

        if "metrics" in self.raw:
            metrics = self.raw["metrics"]
            head = metrics[0:-1] or metrics[0:1]
            text = ", ".join(head)
            if len(metrics) > 1:
                tail = metrics[-1]
                text = text + " and " + tail
        else:
            text = "n/a"

        return text

    @property
    def title(self):
        return self._title or self.description

    @title.setter
    def title(self, value):
        self._title = value

    def metrics(self, *metrics):
        """
        Return a new query with additional metrics.

        ```python
        query.metrics('pageviews', 'page load time')
        ```
        """
        return self.columns("metric", *metrics)

    def dimensions(self, *dimensions):
        """
        Return a new query with additional dimensions.

        ```python
        query.dimensions('search term', 'search depth')
        ```
        """
        return self.columns("dimension", *dimensions)

    @utils.immutable
    def sort(self, *columns, **options):
        """
        Return a new query which will produce results sorted by
        one or more metrics or dimensions. You may use plain
        strings for the columns, or actual `Column`, `Metric`
        and `Dimension` objects.

        Add a minus in front of the metric (either the string or
        the object) to sort in descending order.

        ```python
        # sort using strings
        query.sort('pageviews', '-device type')
        # alternatively, ask for a descending sort in a keyword argument
        query.sort('pageviews', descending=True)

        # sort using metric, dimension or column objects
        pageviews = profile.core.metrics['pageviews']
        query.sort(-pageviews)
        ```
        """

        sorts = self.meta.setdefault("sort", [])

        for column in columns:
            if isinstance(column, Column):
                identifier = column.id
            elif isinstance(column, utils.basestring):
                descending = column.startswith("-") or options.get("descending", False)
                identifier = self.api.columns[column.lstrip("-")].id
            else:
                raise ValueError(
                    "Can only sort on columns or column strings. Received: {}".format(
                        column
                    )
                )

            if descending:
                sign = "-"
            else:
                sign = ""

            sorts.append(sign + identifier)

        self.raw["sort"] = ",".join(sorts)
        return self

    @utils.immutable
    def filter(self, value=None, exclude=False, **selection):
        """Most of the actual functionality lives on the Column
        object and the `all` and `any` functions."""
        filters = self.meta.setdefault("filters", [])

        if value and len(selection):
            raise ValueError(
                "Cannot specify a filter string and a filter keyword selection at the same time."
            )
        elif value:
            value = [value]
        elif len(selection):
            value = select(self.api.columns, selection, invert=exclude)

        filters.append(value)
        self.raw["filters"] = utils.paste(filters, ",", ";")
        return self

    def exclude(self, **selection):
        return self.filter(exclude=True, **selection)

    def build(self, copy=True):
        if copy:
            raw = deepcopy(self.raw)
        else:
            raw = self.raw

        raw["metrics"] = ",".join(self.raw["metrics"])

        if len(raw["dimensions"]):
            raw["dimensions"] = ",".join(self.raw["dimensions"])
        else:
            raw["dimensions"] = None

        return raw

    @property
    def cacheable(self):
        start = "start_date" in self.raw and not utils.date.is_relative(
            self.raw["start_date"]
        )
        end = "end_date" in self.raw and not utils.date.is_relative(
            self.raw["end_date"]
        )
        return start and end

    @property
    def signature(self):
        query = self.build(copy=False)
        standardized_query = sorted(query.items(), key=lambda t: t[0])
        serialized_query = json.dumps(standardized_query)
        return hashlib.sha1(serialized_query.encode("utf-8")).hexdigest()

    def execute(self):
        raw = self.build()

        if self.api.cache and self.cacheable and self.api.cache.exists(self.signature):
            response = self.api.cache.get(raw)
        else:
            try:
                self._wait()
                response = self.endpoint.get(**raw).execute()
            except Exception as err:
                if isinstance(err, TypeError):
                    width = max(map(len, self.raw.keys()))
                    raw = [(key.ljust(width), value) for key, value in self.raw.items()]
                    parameters = utils.paste(raw, "\t", "\n")
                    diagnostics = utils.format(
                        """
                        {message}

                        The query you submitted was:

                        {parameters}
                        """,
                        message=str(err),
                        parameters=parameters,
                    )
                    raise errors.InvalidRequestError(diagnostics)
                else:
                    raise err

        if self.api.cache and self.cacheable:
            self.api.cache.set(raw, response)

        return Report(response, self)

    @property
    def report(self):
        if not self._report:
            self._report = self.get()
        return self._report

    # lazy-loading shortcuts
    def __getattr__(self, name):
        # IPython shell display should not trigger lazy-loading
        # (arguably this is an IPython issue and not our problem, but let's be pragmatic)
        if name == "_ipython_display_":
            raise AttributeError(
                "Query objects have no custom IPython display behavior"
            )
        elif hasattr(self.report, name):
            return getattr(self.report, name)
        else:
            raise AttributeError(
                "'{cls}' object and its associated 'Report' object have no attribute '{name}'".format(
                    cls=self.__class__.__name__,
                    name=name,
                )
            )

    def __repr__(self):
        return "<googleanalytics.query.{} object: {} ({})>".format(
            self.__class__.__name__, self.title, self.profile.name
        )


class CoreQuery(Query):
    """
    CoreQuery is the main way through which to produce reports
    from data in Google Analytics.

    The most important methods are:

    * `metrics` and `dimensions` (both of which you can also pass as
      lists when creating the query)
    * `range` and its shortcuts that have the granularity already set:
      `hourly`, `daily`, `weekly`, `monthly`, `yearly`, `total`
    * `filter` to filter which rows are analyzed before running the query
    * `segment` to filter down to a certain kind of session or user (as
      opposed to `filter` which works on individual rows of data)
    * `limit` to ask for a subset of results
    * `sort` to sort the query

    CoreQuery is mostly immutable: wherever possible, methods
    return a new query rather than modifying the existing one,
    so for example this works as you'd expect it to:

    ```python
    base = profile.query('pageviews')
    january = base.daily('2014-01-01', months=1).get()
    february = base.daily('2014-02-01', months=1).get()
    ```
    """

    # TODO (?)
    # fields
    # userIp / quotaUser
    # https://developers.google.com/analytics/devguides/reporting/core/v3/reference#q_summary

    PRECISION_LEVELS = (
        "FASTER",
        "DEFAULT",
        "HIGHER_PRECISION",
    )
    GRANULARITY_LEVELS = (
        "year",
        "month",
        "week",
        "day",
        "hour",
    )
    GRANULARITY_DIMENSIONS = (
        "ga:year",
        "ga:yearMonth",
        "ga:yearWeek",
        "ga:date",
        "ga:dateHour",
    )

    @utils.immutable
    def precision(self, precision):
        """
        For queries that should run faster, you may specify a lower precision,
        and for those that need to be more precise, a higher precision:

        ```python
        # faster queries
        query.range('2014-01-01', '2014-01-31', precision=0)
        query.range('2014-01-01', '2014-01-31', precision='FASTER')
        # queries with the default level of precision (usually what you want)
        query.range('2014-01-01', '2014-01-31')
        query.range('2014-01-01', '2014-01-31', precision=1)
        query.range('2014-01-01', '2014-01-31', precision='DEFAULT')
        # queries that are more precise
        query.range('2014-01-01', '2014-01-31', precision=2)
        query.range('2014-01-01', '2014-01-31', precision='HIGHER_PRECISION')
        ```
        """

        if isinstance(precision, int):
            precision = self.PRECISION_LEVELS[precision]

        if precision not in self.PRECISION_LEVELS:
            levels = ", ".join(self.PRECISION_LEVELS)
            raise ValueError("Precision should be one of: " + levels)

        if precision != "DEFAULT":
            self.raw.update({"samplingLevel": precision})

        return self

    @utils.immutable
    def interval(self, granularity):
        """
        Note that if you don't specify a granularity (either through the `interval`
        method or through the `hourly`, `daily`, `weekly`, `monthly` or `yearly`
        shortcut methods) you will get only a single result, encompassing the
        entire date range, per metric.
        """

        if granularity == "total":
            return self

        if not isinstance(granularity, int):
            if granularity in self.GRANULARITY_LEVELS:
                granularity = self.GRANULARITY_LEVELS.index(granularity)
            else:
                levels = ", ".join(self.GRANULARITY_LEVELS)
                raise ValueError("Granularity should be one of: lifetime, " + levels)

        dimension = self.GRANULARITY_DIMENSIONS[granularity]
        self.raw["dimensions"].insert(0, dimension)

        return self

    @utils.immutable
    def range(self, start=None, stop=None, months=0, days=0):
        """
        Return a new query that fetches metrics within a certain date range.

        ```python
        query.range('2014-01-01', '2014-06-30')
        ```

        If you don't specify a `stop` argument, the date range will end today. If instead
        you meant to fetch just a single day's results, try:

        ```python
        query.range('2014-01-01', days=1)
        ```

        More generally, you can specify that you'd like a certain number of days,
        starting from a certain date:

        ```python
        query.range('2014-01-01', months=3)
        query.range('2014-01-01', days=28)
        ```

        Note that if you don't specify a granularity (either through the `interval`
        method or through the `hourly`, `daily`, `weekly`, `monthly` or `yearly`
        shortcut methods) you will get only a single result, encompassing the
        entire date range, per metric.

        **Note:** it is currently not possible to easily specify that you'd like
        to query the last last full week(s), month(s) et cetera.
        This will be added sometime in the future.
        """

        start, stop = utils.date.range(start, stop, months, days)

        self.raw.update(
            {
                "start_date": start,
                "end_date": stop,
            }
        )

        return self

    @inspector.implements(range)
    def hourly(self, *vargs, **kwargs):
        return self.interval("hour").range(*vargs, **kwargs)

    @inspector.implements(range)
    def daily(self, *vargs, **kwargs):
        """
        Return a new query that fetches metrics within a certain date
        range, summarized by day. This method is identical to
        `CoreQuery#range` but it sets the default granularity to
        `granularity='day'`.
        """
        return self.interval("day").range(*vargs, **kwargs)

    @inspector.implements(range)
    def weekly(self, *vargs, **kwargs):
        """
        Return a new query that fetches metrics within a certain date
        range, summarized by week. This method is identical to
        `CoreQuery#range` but it sets the default granularity to
        `granularity='week'`.
        """
        return self.interval("week").range(*vargs, **kwargs)

    @inspector.implements(range)
    def monthly(self, *vargs, **kwargs):
        """
        Return a new query that fetches metrics within a certain date
        range, summarized by month. This method is identical to
        `CoreQuery#range` but it sets the default granularity to
        `granularity='month'`.
        """
        return self.interval("month").range(*vargs, **kwargs)

    @inspector.implements(range)
    def yearly(self, *vargs, **kwargs):
        """
        Return a new query that fetches metrics within a certain date
        range, summarized by year. This method is identical to
        `CoreQuery#range` but it sets the default granularity to
        `granularity='year'`.
        """
        return self.interval("year").range(*vargs, **kwargs)

    @inspector.implements(range)
    def total(self, *vargs, **kwargs):
        return self.range(*vargs, **kwargs)

    @utils.immutable
    def step(self, maximum):
        """
        Return a new query with a maximum amount of results to be returned
        in any one request, without implying that we should stop
        fetching beyond that limit (unlike `CoreQuery#limit`.)

        Useful in debugging pagination functionality.

        Perhaps also useful when you  want to be able to decide whether to
        continue fetching data, based  on the data you've already received.
        """

        self.raw["max_results"] = maximum
        return self

    @utils.immutable
    def limit(self, *_range):
        """
        Return a new query, limited to a certain number of results.

        ```python
        # first 100
        query.limit(100)
        # 50 to 60
        query.limit(50, 10)
        ```

        Please note carefully that Google Analytics uses
        1-indexing on its rows.
        """

        # uses the same argument order as
        # LIMIT in a SQL database
        if len(_range) == 2:
            start, maximum = _range
        else:
            start = 1
            maximum = _range[0]

        self.meta["limit"] = maximum

        self.raw.update(
            {
                "start_index": start,
                "max_results": maximum,
            }
        )
        return self

    @utils.immutable
    def segment_sequence(
        self, followed_by=False, immediately_followed_by=False, first=False
    ):
        # sequences are just really hard to "simplify" because so much is possible

        if followed_by or immediately_followed_by:
            method = "sequence"
        else:
            method = "condition"

        raise NotImplementedError()

    @utils.immutable
    def segment(self, value=None, scope=None, metric_scope=None, **selection):
        """
        Return a new query, limited to a segment of all users or sessions.

        Accepts segment objects, filtered segment objects and segment names:

        ```python
        query.segment(account.segments['browser'])
        query.segment('browser')
        query.segment(account.segments['browser'].any('Chrome', 'Firefox'))
        ```

        Segment can also accept a segment expression when you pass
        in a `type` argument. The type argument can be either `users`
        or `sessions`. This is pretty close to the metal.

        ```python
        # will be translated into `users::condition::perUser::ga:sessions>10`
        query.segment('condition::perUser::ga:sessions>10', type='users')
        ```

        See the [Google Analytics dynamic segments documentation][segments]

        You can also use the `any`, `all`, `followed_by` and
        `immediately_followed_by` functions in this module to
        chain together segments.

        Everything about how segments get handled is still in flux.
        Feel free to propose ideas for a nicer interface on
        the [GitHub issues page][issues]

        [segments]: https://developers.google.com/analytics/devguides/reporting/core/v3/segments#reference
        [issues]: https://github.com/debrouwere/google-analytics/issues
        """

        """
        Technical note to self about segments:

        * users or sessions
        * sequence or condition
        * scope (perHit, perSession, perUser -- gte primary scope)

        Multiple conditions can be ANDed or ORed together; these two are equivalent

            users::condition::ga:revenue>10;ga:sessionDuration>60
            users::condition::ga:revenue>10;users::condition::ga:sessionDuration>60

        For sequences, prepending ^ means the first part of the sequence has to match
        the first session/hit/...

        * users and sessions conditions can be combined (but only with AND)
        * sequences and conditions can also be combined (but only with AND)

        sessions::sequence::ga:browser==Chrome;
        condition::perHit::ga:timeOnPage>5
        ->>
        ga:deviceCategory==mobile;ga:revenue>10;

        users::sequence::ga:deviceCategory==desktop
        ->>
        ga:deviceCategory=mobile;
        ga:revenue>100;
        condition::ga:browser==Chrome

        Problem: keyword arguments are passed as a dictionary, not an ordered dictionary!
        So e.g. this is risky

            query.sessions(time_on_page__gt=5, device_category='mobile', followed_by=True)
        """

        SCOPES = {
            "hits": "perHit",
            "sessions": "perSession",
            "users": "perUser",
        }
        segments = self.meta.setdefault("segments", [])

        if value and len(selection):
            raise ValueError(
                "Cannot specify a filter string and a filter keyword selection at the same time."
            )
        elif value:
            value = [self.api.segments.serialize(value)]
        elif len(selection):
            if not scope:
                raise ValueError("Scope is required. Choose from: users, sessions.")

            if metric_scope:
                metric_scope = SCOPES[metric_scope]

            value = select(self.api.columns, selection)
            value = [
                [scope, "condition", metric_scope, condition] for condition in value
            ]
            value = ["::".join(filter(None, condition)) for condition in value]

        segments.append(value)
        self.raw["segment"] = utils.paste(segments, ",", ";")
        return self

    def users(self, **kwargs):
        return self.segment(scope="users", **kwargs)

    def sessions(self, **kwargs):
        return self.segment(scope="sessions", **kwargs)

    @utils.immutable
    def next(self):
        """
        Return a new query with a modified `start_index`.
        Mainly used internally to paginate through results.
        """
        step = self.raw.get("max_results", 1000)
        start = self.raw.get("start_index", 1) + step
        self.raw["start_index"] = start
        return self

    def get(self):
        """
        Run the query and return a `Report`.

        This method transparently handles paginated results, so even for results that
        are larger than the maximum amount of rows the Google Analytics API will
        return in a single request, or larger than the amount of rows as specified
        through `CoreQuery#step`, `get` will leaf through all pages,
        concatenate the results and produce a single Report instance.
        """

        cursor = self
        report = None
        is_complete = False
        is_enough = False

        while not (is_enough or is_complete):
            chunk = cursor.execute()

            if report:
                report.append(chunk.raw[0], cursor)
            else:
                report = chunk

            is_enough = len(report.rows) >= self.meta.get("limit", float("inf"))
            is_complete = chunk.is_complete
            cursor = cursor.next()

        return report


class RealTimeQuery(Query):
    """
    A query against the [Google Analytics Real Time API][realtime].

    **Note:** brand new! Please test and submit any issues to GitHub.

    [realtime]: https://developers.google.com/analytics/devguides/reporting/realtime/v3/reference/data/realtime#resource
    """

    @property
    def endpoint(self):
        return self.account.service.data().realtime()

    @utils.immutable
    def limit(self, maximum):
        """
        Return a new query, limited to a certain number of results.

        Unlike core reporting queries, you cannot specify a starting
        point for live queries, just the maximum results returned.

        ```python
        # first 50
        query.limit(50)
        ```
        """

        self.meta["limit"] = maximum
        self.raw.update(
            {
                "max_results": maximum,
            }
        )
        return self

    def get(self):
        return self.execute()


# TODO: consider moving the blueprint functionality to a separate Python package


def describe(profile, description):
    """
    Generate a query by describing it as a series of actions
    and parameters to those actions. These map directly
    to Query methods and arguments to those methods.

    This is an alternative to the chaining interface.
    Mostly useful if you'd like to put your queries
    in a file, rather than in Python code.
    """
    api_type = description.pop("type", "core")
    api = getattr(profile, api_type)
    return refine(api.query, description)


def refine(query, description):
    """
    Refine a query from a dictionary of parameters that describes it.
    See `describe` for more information.
    """

    for attribute, arguments in description.items():
        if hasattr(query, attribute):
            attribute = getattr(query, attribute)
        else:
            raise ValueError("Unknown query method: " + attribute)

        # query descriptions are often automatically generated, and
        # may include empty calls, which we skip
        if utils.isempty(arguments):
            continue

        if callable(attribute):
            method = attribute
            if isinstance(arguments, dict):
                query = method(**arguments)
            elif isinstance(arguments, list):
                query = method(*arguments)
            else:
                query = method(arguments)
        else:
            setattr(attribute, arguments)

    return query



import datetime
import re

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta


# Python 2 and 3 compatibility
try:
    basestring = basestring
except NameError:
    basestring = str


def serialize(value):
    if isinstance(value, datetime.date):
        return value.isoformat()
    else:
        return value


def extract(obj):
    if isinstance(obj, datetime.date):
        if hasattr(obj, "date"):
            return obj.date()
        else:
            return obj
    else:
        raise ValueError(
            "Can only extract date for type: date, datetime. Received: {}".format(obj)
        )


def parse_description(s):
    today = datetime.date.today()
    if s == "today":
        return today
    elif s == "yesterday":
        return today - relativedelta(days=1)
    else:
        match = re.match("(\d+)daysAgo", s)
        if match:
            return today - relativedelta(days=int(match.group(1)))
        else:
            raise ValueError(
                "Can only parse descriptions of the format: today, yesterday, ndaysAgo"
            )


def normalize(obj):
    if obj == None:
        return None
    elif isinstance(obj, datetime.date):
        return extract(obj)
    elif isinstance(obj, basestring):
        try:
            return extract(parse(obj))
        except ValueError:
            try:
                return extract(parse_description(obj))
            except ValueError:
                raise ValueError("Cannot parse date or description: " + obj)
    else:
        raise ValueError(
            "Can only normalize dates of type: date, datetime, basestring."
        )


def range(start=None, stop=None, months=0, days=0):
    yesterday = datetime.date.today() - relativedelta(days=1)
    start = normalize(start) or yesterday
    stop = normalize(stop)
    is_past = days < 0 or months < 0

    if days or months:
        if start and stop:
            raise Exception(
                "A daterange cannot be defined using stop alongside months or days."
            )
        else:
            if is_past:
                days = days + 1
            else:
                days = days - 1

            delta = relativedelta(days=days, months=months)

            stop = start + delta

    stop = stop or start
    return map(serialize, sorted((start, stop)))


def is_relative(datestring):
    return not "-" in datestring



#! /usr/bin/env python
# coding: utf-8
#

"""ForecastGA: Logging Utility"""

import logging
from logging import DEBUG, INFO, ERROR, Formatter, getLogger

# file output
FILE_HANDLER = logging.FileHandler(filename="forecastga.error.log")

FILE_HANDLER.setFormatter(
    Formatter("%(asctime)s [%(levelname)s]" "  %(name)s,%(lineno)s  %(message)s")
)
FILE_HANDLER.setLevel(DEBUG)

# console output
CONSOLE_HANDLER = logging.StreamHandler()
CONSOLE_HANDLER.setLevel(ERROR)
CONSOLE_HANDLER.setFormatter(Formatter("%(message)s"))

SDCT_LOGGER = getLogger("forecastga")

# add handlers
SDCT_LOGGER.addHandler(CONSOLE_HANDLER)
SDCT_LOGGER.addHandler(FILE_HANDLER)
SDCT_LOGGER.setLevel(DEBUG)

logging.captureWarnings(True)


def get_logger(log_name, level=INFO):
    """
    :param level:   CRITICAL = 50
                    ERROR = 40
                    WARNING = 30
                    INFO = 20
                    DEBUG = 10
                    NOTSET = 0
    :type log_name: str
    :type level: int
    """
    module_logger = SDCT_LOGGER.getChild(log_name)
    if level:
        module_logger.setLevel(level)
    return module_logger
